"""

Author: Petrou Dimitrios 
Organization: Athena Research Center
Project Name:  STELAR EU 
Project Info: https://stelar-project.eu/

"""
# from typing import Any

from urllib.parse import urlparse

from minio.datatypes import Object

S3ObjSpec = str | tuple[str, str] | Object


def is_s3url(url: str) -> bool:
    """Check if a given URL is an S3 URL.

    Args:
        url (str): The URL to check.
    Returns:
        bool: True if the URL is an S3 URL, False otherwise.
    """
    u = urlparse(url)
    return u.scheme == "s3"


def s3spec_to_pair(objspec: S3ObjSpec) -> tuple[str, str]:
    """Converts an S3 object specification to a pair of bucket and object name.

    Args:
        objspec (S3ObjSpec): A URL (string), tuple of two strings,
        or minio.datatypes.Object instance representing an S3 object.
    Returns:
        tuple[str, str]: A pair of strings, representing the bucket and object name.
    """
    if isinstance(objspec, str):
        u = urlparse(objspec)
        if u.scheme != "s3":
            raise ValueError(f"Invalid S3 object url: {objspec}")
        bucket = u.netloc
        obj = u.path.lstrip("/")
        return bucket, obj
    elif isinstance(objspec, tuple) and len(objspec) == 2:
        return objspec
    elif isinstance(objspec, Object):
        return objspec.bucket_name, objspec.object_name
    else:
        raise TypeError(f"Invalid type S3 object spec: {repr(objspec)}")


def s3spec_to_dict(objspec: S3ObjSpec) -> dict[str, str]:
    """Converts an S3 object specification to a dictionary with bucket and object keys.

    The main use case for this function is to call client.s3 methods quickly:

    .. code-block:: python

        client.s3.get_object(**s3spec_to_dict("s3://bucket/object"))

    Args:
        objspec (S3ObjSpec): A URL (string), tuple of two strings,
        or minio.datatypes.Object instance representing an S3 object.
    Returns:
        dict[str, str]: A dictionary with keys "bucket_name" and "object_name".
    """
    bucket, obj = s3spec_to_pair(objspec)
    return {"bucket_name": bucket, "object_name": obj}


def s3spec_to_url(objspec: S3ObjSpec) -> str:
    """Converts an S3 object specification to a URL string.

    Args:
        objspec (S3ObjSpec): A URL (string), tuple of two strings,
        or minio.datatypes.Object instance representing an S3 object.
    Returns:
        str: A URL string representing the S3 object.
    """
    bucket, obj = s3spec_to_pair(objspec)
    return f"s3://{bucket}/{obj}"



import pytest
from proxy_utils import ProxyTestObj

from stelar.client.proxy import Property, Proxy, Registry, RegistryCatalog


def test_create_catalog():
    class Foo(ProxyTestObj):
        a = Property()
        data = {"a": 0}

    class Bar(ProxyTestObj):
        a = Property()
        data = {"a": 0}

    c = RegistryCatalog()
    rFoo = Registry(c, Foo)

    assert c.registry_for(Foo) is rFoo
    with pytest.raises(KeyError):
        c.registry_for(Bar)



import pytest


@pytest.fixture()
def vocab1(testcli):
    c = testcli
    vocab1 = c.vocabularies.create(name="vocab1", tags=["tag1", "tag2", "tag3"])
    yield vocab1
    vocab1.delete()


def test_vocabulary_create(testcli, vocab1):
    # c = testcli
    assert vocab1.name == "vocab1"
    assert vocab1.tagnames == ["tag1", "tag2", "tag3"]
    assert vocab1.tagspecs == ["vocab1:tag1", "vocab1:tag2", "vocab1:tag3"]


def test_vocabulary_list(testcli):
    c = testcli
    oldvoc = set(c.vocabularies[:])
    vocab1 = c.vocabularies.create(name="vocab1")
    vocab2 = c.vocabularies.create(name="vocab2")
    assert set(c.vocabularies[:]) == oldvoc | {vocab1, vocab2}
    vocab1.delete()
    vocab2.delete()
    assert set(c.vocabularies[:]) == oldvoc



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type, TypeVar

from .property import Property
from .proxy import Proxy
from .proxylist import ProxyList
from .refs import Reference

if TYPE_CHECKING:
    from ..client import Client
    from .property import RefList
    from .registry import Registry
    from .schema import Schema

ProxyClass = TypeVar("ProxyClass", bound=Proxy)


class ProxySynclist:
    """A container for the proxies that need to be sync'd after
    an operation.
    """

    def __init__(self, l: list[Proxy] = []):
        self.tosync = list()
        for a in l:
            self.add(a)

    def add(self, ref: Proxy | ProxyList):
        """Add a proxy or the elements of a proxy list to the synclist.

        Any other types are ignored.
        """
        if isinstance(ref, ProxyList):
            for p in ref:
                self.add(p)
        elif isinstance(ref, Proxy):
            self.tosync.append(ref)

    def trigger_properties(self, schema: Schema):
        for p in schema.properties.values():
            if isinstance(p, Reference) and p.trigger_sync:
                yield p

    def on_create(self, proxy_type: Type[ProxyClass], properties: dict[str, Any]):
        for p in self.trigger_properties(proxy_type.proxy_schema):
            self.add(properties.get(p.name))

    def on_create_proxy(self, proxy: ProxyClass):
        for p in self.trigger_properties(proxy.proxy_schema):
            self.add(p.get(proxy))

    def on_delete(self, proxy: Proxy):
        for p in proxy.proxy_schema.properties.values():
            # Add all present properties referenced from this property
            self.add(getattr(proxy, p.name, None))

    def on_update(self, proxy: Proxy, prop: Property, newref: Proxy):
        if isinstance(prop, Reference) and prop.trigger_sync:
            oldref = getattr(proxy, prop.name)
            self.add(oldref)
            self.add(newref)

    def on_update_all(self, proxy: ProxyClass):
        for p in self.trigger_properties(proxy.proxy_schema):
            self.add(p.get(proxy))

    def sync(self):
        for prx in self.tosync:
            prx.proxy_sync()
        self.tosync.clear()



from __future__ import annotations

import copy
from collections.abc import MutableMapping, MutableSequence
from io import StringIO
from typing import TYPE_CHECKING, Any

from .exceptions import EntityError
from .fieldvalidation import AnyField, DictField, ListField, NameField, UUIDField
from .proxy import Proxy

if TYPE_CHECKING:
    pass
    # from ..client import Client

Entity = dict[str, Any]


class Property:
    """A Python descriptor for implementing access and updating of
    fields of proxy objects.

    A property object performs the following roles:
     -   Is a python 'descriptor' (implements getter, setter and deleter) for
         proxy classes
     -   Holds a number of metadata that determine the behaviour of the proxy
         (validation, conversion to entity, nullabilty, updatability, optionality,
         view, default values at creation, etc)
     -   Performs internalization/externalization for the data.

    Properties are the mechanism for most of the functionality of the
    STELAR client.
    """

    def __init__(
        self,
        *,
        validator=None,
        updatable=False,
        optional=False,
        entity_name=None,
        doc=None,
        create_default=None,
        short=None,
    ):
        """Constructs a proxy property descriptor

        Args:
            validator (FieldValidator)
            updatable (bool): If false, property cannot be set
            optional (bool): If false, property cannot be deleted
            entity_name (str|None): Corresponds to the entity field name. If not given,
                the same as the property name.
            doc (str|None): A piece of text that describes the property. This is
                used to form the full documentation for the property.
            create_default (Any): If provided, it specifies an attribute on the entity's registry
                (typically implemented as a 'functools.cached_property') that is used to initialize
                new entities, when the user does not provide a value for this property.
                Note: this is a *proxy_attr value* that must be convertible to an entity value.
            short (bool|None): Denotes whether this property is included in "short presentations"
                of the entity. If None, a heuristic based on the name and type is used.
        """
        self.updatable = updatable
        self.isId = False
        self.isName = False
        self.isExtras = False
        self.optional = optional
        if validator is None:
            self.validator = AnyField()
        elif isinstance(validator, type):
            self.validator = validator()
        else:
            self.validator = validator
        self.entity_name = entity_name
        self.owner = self.name = None
        self.create_default = create_default
        self.short = short
        self.doc = doc

    def autodoc(self, doc, repr_type, repr_constraints):
        INDENT = "    "
        out = StringIO()

        typespec = str(repr_type)
        constraints = list(repr_constraints)

        if "nullable" in repr_constraints:
            typespec += "|None"
            constraints.remove("nullable")

        if doc is None:
            doc = f"The '{self.name}' field"

        f = [typespec]

        fp = []
        if not self.updatable:
            fp.append("read-only")
        if self.optional:
            fp.append("deletable")
        if fp:
            f.append(f"({', '.join(fp)}):")
        else:
            f.append(":")

        f += [doc, *constraints]

        if self.entity_name != self.name:
            f += ["JSON field:", self.entity_name]

        print(INDENT, *f, file=out)
        return out.getvalue()

    @property
    def qualname(self):
        return f"{self.owner.__name__}.{self.name}"

    def __repr__(self):
        return f"<Property {self.qualname}>"

    def __str__(self):
        return self.qualname

    def __set_name__(self, owner, name):
        if not issubclass(owner, Proxy):
            raise TypeError(f"Class {owner.__qualname__} must inherit from class Proxy")
        self.owner = owner
        self.name = name
        if self.entity_name is None:
            self.entity_name = name
        self.__doc__ = self.autodoc(
            self.doc, self.validator.repr_type(), self.validator.repr_constraints()
        )

    # def check_value(self, value):
    #    if value is ... and not self.optional:
    #        raise ValueError(f"Property '{self.name}' is not optional")
    #    return value

    def missing(self, *, proxy=None, registry=None, **kwargs):
        """Provides missing values during entity creation.

        Args:
        proxy (a 'create proxy' is ok) or registry, one muste be provided.
        """
        if registry is None:
            registry = proxy.proxy_registry
        if self.create_default:
            return getattr(registry, self.create_default, ...)
        if hasattr(self.validator, "default"):
            return self.validator.default
        return ...

    def get(self, obj):
        """Low-level getter"""
        if obj.proxy_attr is None:
            obj.proxy_sync()
        return obj.proxy_attr[self.name]

    def touch(self, obj):
        """Transition the initial value of a clean proxy to the
        'proxy_changed' dictionary.

        This is done only on the first update to an attribute,
        in order to allow for the proxy_reset() functionality.

        Args:
            obj (Proxy): The proxy object whose property is touched.
        Returns:
            True if the touch actually happened, False if the property
            was already touched.
        """
        if obj.proxy_changed is None:
            # Initialize proxy_changed on clean object
            if obj.proxy_attr is None:
                obj.proxy_sync()
            obj.proxy_changed = {self.name: obj.proxy_attr[self.name]}
            return True
        elif self.name not in obj.proxy_changed:
            # Record only first change
            obj.proxy_changed[self.name] = obj.proxy_attr[self.name]
            return True
        return False

    def validate(self, obj, value):
        return self.validator.validate(value)

    def set(self, obj, value):
        """Low-level setter"""
        value = self.validate(obj, value)
        self.touch(obj)
        # update the value
        obj.proxy_attr[self.name] = value

    def convert_entity_to_proxy(self, proxy: Proxy, entity: Any, **kwargs):
        """Update proxy dict to represent this property from the entity"""
        if self.optional:
            entity_value = entity.get(self.entity_name, ...)
        else:
            try:
                entity_value = entity[self.entity_name]
            except KeyError as e:
                raise EntityError(
                    f"Entity does not have attribute {self.entity_name}"
                ) from e

        try:
            value = (
                self.validator.convert_to_proxy(entity_value, **kwargs)
                if entity_value is not None
                else None
            )
        except Exception as e:
            raise EntityError(
                f"Error converting entity '{self.name}' value to proxy"
            ) from e

        proxy.proxy_attr[self.name] = value

    def convert_proxy_to_entity(self, proxy: Proxy, entity: dict, **kwargs):
        """Update entity dict to represent this property from the proxy.
        The `changes` flag is true when the proxy_attr is actually the Proxy.proxy_changes
        dict.
        """
        proxy_value = proxy.proxy_attr[self.name]
        if proxy_value is ...:
            return
        if proxy_value is None:
            entity[self.entity_name] = None
        else:
            entity[self.entity_name] = self.validator.convert_to_entity(
                proxy_value, **kwargs
            )

    def convert_to_create(
        self, proxy_type: type, create_props: Entity, entity_props: Entity, **kwargs
    ):
        """Convert a value to be used for entity creation.

        Args:
            client (Client): The client used for creation.
            proxy_type (ProxyClass): The entity type being created.
            create_props: The object passed to the create client call
            entity_props: The entity object given to the create API call.
        """
        if self.name not in create_props:
            defval = self.missing(**kwargs)
            if defval is not ...:
                entity_props[self.entity_name] = self.validator.convert_to_entity(
                    defval, **kwargs
                )
            return

        proxy_value = create_props[self.name]
        proxy_value = self.validator.validate(proxy_value, **kwargs)
        if proxy_value is None:
            entity_value = None
        else:
            entity_value = self.validator.convert_to_entity(proxy_value, **kwargs)
        entity_props[self.entity_name] = entity_value

    def __get__(self, obj, objtype=None):
        val = self.get(obj)

        # The attribute is deleted
        if val is ...:
            raise AttributeError(f"Property '{self.name}' is not present")
        return val

    def __set__(self, obj, value):
        if not self.updatable:
            raise AttributeError(f"Property '{self.name}' is read-only")
        if value is ...:
            raise ValueError("Properties cannot be set to '...'")
        self.set(obj, value)
        obj.proxy_autocommit()

    def __delete__(self, obj):
        if not self.optional:
            raise AttributeError(f"Property '{self.name}' is not optional")
        if obj.proxy_attr is None:
            obj.proxy_sync()
        if obj.proxy_attr[self.name] is ...:
            raise AttributeError(f"{self.name} is not present")
        self.set(obj, ...)


class Id(Property):
    """A Python descriptor for implementing entity ID access."""

    def __init__(self, entity_name=None, doc=None):
        if doc is None:
            doc = "The ID"
        super().__init__(
            validator=UUIDField(nullable=False), entity_name=entity_name, doc=doc
        )
        self.isId = True

    def __get__(self, obj, objtype=None):
        return obj.proxy_id

    def __set__(self, obj, value):
        raise AttributeError("Entity ID attribute cannot be set")

    def __delete__(self, obj):
        raise AttributeError("Entity ID attribute cannot be deleted")


class NameId(Property):
    """Many entities have a name field with a unique constaint."""

    def __init__(self, entity_name=None, doc=None, validator=NameField):
        if doc is None:
            doc = "The name field, which is a unique string identifying the entity."
        super().__init__(validator=validator, entity_name=entity_name, doc=doc)
        self.isName = True


# -----------------------------------------
# Properties returning proxies
#
# These properties are used to implement properties that return
# proxy objects, instead of simple values. This is used for dict-valued
# and list-valued properties. The proxy object is responsible for
# updating the entity when the value is changed.
#
# Implementation:
# Suppose that property 'inputs' is a dict-valued property. The property getter
# __get__() returns a DictPropertyProxy object, which is a MutableMapping.
#
# For a mutating operation (__setitem__ or __delitem__), the DictPropertyProxy
# object updates the dictionary and then sets the property value to the new
# updated dictionary. Note that the whole new dictionary is set (and auto-updated)
# not just one entry.
#
# For a non-mutating operation (__getitem__ or __len__ or __iter__),
# the DictPropertyProxy object simply returns the
# value from the dictionary.
# -----------------------------------------


class PropertyProxy:
    def __init__(self, proxy, property):
        self._proxy = proxy
        self._property = property
        self._name = property.name


class ProxyProperty(Property):
    """A proxy property allows for properties whose values are objects.

    As an example, a dict-valued property can be accessed via a proxy object,
    which detects changes to the dictionary and automatically updates the
    entity proxy.

    For this to work, the proxy property is parametrized with a PropertyProxy
    class, whose instances are to be returned by the __get__ method.
    """

    def __init__(self, property_proxy_class, **kwargs):
        super().__init__(**kwargs)
        self.property_proxy_class = property_proxy_class

    def __get__(self, obj, objtype=None):
        val = super().__get__(obj, objtype)
        if val is None:
            return None
        else:
            return self.property_proxy_class(obj, self)


class DictProxy(PropertyProxy, MutableMapping):
    def __getitem__(self, key):
        return self._proxy.proxy_attr[self._name][key]

    def __setitem__(self, key, value):
        newval = copy.copy(self._proxy.proxy_attr[self._name])
        newval[key] = value
        setattr(self._proxy, self._name, newval)

    def __delitem__(self, key):
        newval = copy.copy(self._proxy.proxy_attr[self._name])
        del newval[key]
        setattr(self._proxy, self._name, newval)

    def __iter__(self):
        return iter(self._proxy.proxy_attr[self._name])

    def __len__(self):
        return len(self._proxy.proxy_attr[self._name])

    def __repr__(self):
        return f"<DictProxy {self._name}: {self._proxy.proxy_attr[self._name]}>"

    def __str__(self):
        return str(self._proxy.proxy_attr[self._name])

    def __or__(self, other):
        return self._proxy.proxy_attr[self._name] | other

    def __ior__(self, other):
        newval = copy.copy(self._proxy.proxy_attr[self._name])
        newval |= other
        setattr(self._proxy, self._name, newval)
        return self


class DictProperty(ProxyProperty):
    def __init__(self, key_type, value_type, *, nullable=False, **kwargs):
        super().__init__(
            DictProxy,
            validator=DictField(key_type, value_type, nullable=nullable),
            **kwargs,
        )


class ListProxy(PropertyProxy, MutableSequence):
    # __getitem__, __setitem__, __delitem__, __len__, insert

    def __getitem__(self, index):
        return self._proxy.proxy_attr[self._name][index]

    def __setitem__(self, index, value):
        newval = copy.copy(self._proxy.proxy_attr[self._name])
        newval[index] = value
        setattr(self._proxy, self._name, newval)

    def __delitem__(self, index):
        newval = copy.copy(self._proxy.proxy_attr[self._name])
        del newval[index]
        setattr(self._proxy, self._name, newval)

    def __len__(self):
        return len(self._proxy.proxy_attr[self._name])

    def insert(self, index, value):
        newval = copy.copy(self._proxy.proxy_attr[self._name])
        newval.insert(index, value)
        setattr(self._proxy, self._name, newval)

    def __repr__(self):
        return f"<ListProxy {self._name}: {self._proxy.proxy_attr[self._name]}>"

    def __str__(self):
        return str(self._proxy.proxy_attr[self._name])


class ListProperty(ProxyProperty):
    def __init__(self, value_type, *, nullable=False, **kwargs):
        super().__init__(
            ListProxy,
            validator=ListField(value_type, nullable=nullable),
            **kwargs,
        )



from __future__ import annotations

import re
from enum import Enum


class ProxyState(Enum):
    ERROR = 0  # means that a proxy is no longer valid (maybe the entity was deleted?)
    EMPTY = 1  # means that a proxy contains no entity data
    CLEAN = 2  # means that a proxy contains entity data but no changes since last sync
    DIRTY = 3  # means that a proxy contains changes since last sync


# -------------------------
# Tag-related declarations
# -------------------------
TAGNAME_PATTERN = re.compile(r"[A-Za-z0-9 _-]{2,100}")
TAGSPEC_PATTERN = re.compile(r"((.{2,100})\:)?([A-Za-z0-9 _-]{2,100})")


def validate_tagname(tagname: str) -> bool:
    """Check if a string is formatted correctly as a tagname"""
    return TAGNAME_PATTERN.fullmatch(tagname) is not None


def validate_tagspec(tagspec: str) -> bool:
    """Check if a string is formatted correctly as a tagspec"""
    return TAGSPEC_PATTERN.fullmatch(tagspec) is not None


def tag_split(tagspec: str) -> tuple[str | None, str]:
    """Split a tagspec into a pair or (<vocabulary-name> , <tag-name>).

    Properly, a tagspec is either <tag-name>  or <vocabulary-name>:<tag-name>,
    where

    <tagname>
        is a string made only of lower-case alphanumerics, hyphen (-) and underscore (_),
        and of length in [2,100]
    <vocabulary-name>
        is any string (which may contain spaces and other ascii characters) of
        length [2,100].

    Arguments
    ---------
        tagspec (str): the tag specification
    Returns
    -------
        (str|None, str): a pair of (<vocabulary-name> , <tag-name>)
    Raises
    ------
        ValueError: if the tagspec is not valid
    """
    m = TAGSPEC_PATTERN.fullmatch(tagspec)
    if m is None:
        raise ValueError("Invalid tag specification")
    return m.groups()[1:]


def tag_join(vocname: str | None, tagname: str) -> str:
    """Return the tagspec of a tag and a vocabulary.

    Args:
        vocname  (str|None): the vocabulary name
        tagname: the tag name
    Returns:
        (str) the tagspec
    """
    return f"{vocname}:{tagname}" if vocname is not None else tagname



from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar

from .property import Id, NameId, Property
from .proxy import Proxy

# ----------------------------------------------------------
#  Proxy Schema
#
#  A collection of attributes defined for an entity.
#  Each attribute is an instance of Property and
#  contains metadata related to this attribute.
#
#  Subclasses of Proxy are either entities or 'abstract'.
#  Abstract subclasses are not instantiated directly.
#  Entities cannot inherit other entities. However, they
#  can inherit abstract classes, in which case they
#  inherit copies of their properties.
#
# ----------------------------------------------------------


def inheritance_for_entity_class(cls: type):
    # Helper for schema initialization
    # This function collects properties and
    # base classes for a new entity class

    abstract_bases = []
    inherited_properties = {
        name: prop for name, prop in cls.__dict__.items() if isinstance(prop, Property)
    }

    for sc in cls.__mro__[1:]:
        if issubclass(sc, Proxy):
            # The class must be abstract
            if hasattr(sc, "proxy_schema"):
                raise TypeError(
                    f"Entity {cls.__name__} inherits from entity {sc.__name__}"
                )
            abstract_bases.append(sc)
            for name, prop in sc.__dict__.items():
                if isinstance(prop, Property) and name not in inherited_properties:
                    inherited_properties[name] = prop

    return abstract_bases, inherited_properties


class Schema:
    """A class that holds all information related to an entity proxy class.
    This includes the list of proxied properties, and other information
    """

    # Declare attributes
    properties: dict[str, Property]

    # Class attribute, registers schemas for entity names.
    entity_schema: dict[str, Schema] = dict()

    @classmethod
    def for_entity(cls, name: str) -> Schema:
        """Return the schema for the given named entity

        Args:
            name (str): the name of an entity (e.g. 'Dataset', 'Group', etc)
        Returns:
            the schema for the give entity, or None.
        """
        return cls.entity_schema.get(name)

    @property
    def class_name(self):
        """Return the class name for the"""
        return self.cls.__name__

    def __init__(self, cls):
        """The proxy class for this schema."""
        self.cls = cls
        self.properties = {}
        self.id = None
        self.name_id = None
        self.extras = None
        self.all_fields = dict()
        self.all_entity_fields = set()

        # Initialize the properties list
        # N.B. This does not check the superclasses
        self.abstract_base_classes, inherited_properties = inheritance_for_entity_class(
            cls
        )
        # print("Class", cls, "has properties", inherited_properties)

        for name, prop in inherited_properties.items():
            if prop.isId:
                assert isinstance(prop, Id)
                if self.id is not None:
                    raise TypeError(
                        f"Multiple ID attributes defined for {cls.__qualname__}"
                    )
                self.id = prop
            elif prop.isName:
                assert isinstance(prop, NameId)
                if self.name_id is not None:
                    raise TypeError(
                        f"Multiple nameID attributes defined for {cls.__qualname__}"
                    )
                self.name_id = prop
                # Name is added to the properties !
                self.properties[name] = prop
            elif prop.isExtras:
                if self.extras is not None:
                    raise TypeError(
                        f"Multiple Extras attributes defined for {cls.__qualname__}"
                    )
                self.extras = prop
                # Extras is added to the properties !
                self.properties[name] = prop
            else:
                self.properties[name] = prop

            # Add the name to all_fields
            self.all_fields[name] = prop
            self.all_entity_fields.add(prop.entity_name)

        # check that we have specified an ID attribute, and add a default
        # ID attribute otherwise
        if self.id is None:
            # check that attribute 'id' is available!
            if hasattr(cls, "id") or "id" in self.properties:
                raise TypeError(
                    f"A member named 'id' is present but it is not marked as entity ID"
                )

            self.id = Id()
            cls.id = self.id
            self.id.__set_name__(cls, "id")

        # Register yourself
        self.entity_schema[self.class_name] = self

    def get_id(self, entity) -> str:
        """Return the entity ID from the entity object"""
        return entity[self.id.entity_name]

    SHORT_NAMES = {
        "id",
        "name",
        "title",
        "author",
        "maintainer",
        "resource_type",
        "url",
    }

    def short_list(self, additional={}):
        """Return a 'short list' of important field names.

        The main purpose for this is to provide a reduced set of names to
        use in visual representations of proxy instances.
        """
        return [self.id.name] + [
            name
            for name, prop in self.properties.items()
            if (
                prop.short is True
                or name in additional
                or (prop.short is None and name in self.SHORT_NAMES)
            )
        ]

    @staticmethod
    def check_non_entity(cls):
        # Nothing yet!
        pass



from __future__ import annotations

"""Classes used to access the STELAR API.
"""
from typing import TYPE_CHECKING, Optional

from .proxy import EntityNotFound, Proxy, ProxyCursor, ProxyList, ProxyOperationError
from .utils import client_for

if TYPE_CHECKING:
    from .client import Client

    APIContext = Proxy | Client | ProxyCursor | ProxyList


class api_context:
    def __init__(self, arg: APIContext):
        from .client import Client

        if isinstance(arg, Proxy):
            self.proxy = arg
            self.client = client_for(self.proxy)
            self.proxy_id = self.proxy.proxy_id
            self.proxy_type = type(self.proxy)
        elif isinstance(arg, (ProxyCursor, ProxyList)):
            self.proxy = None
            self.client = arg.client
            self.proxy_id = None
            self.proxy_type = arg.proxy_type
        elif isinstance(arg, Client):
            self.proxy = None
            self.client = arg
            self.proxy_id = None
            self.proxy_type = None


class api_model:
    @staticmethod
    def from_value(value):
        if isinstance(value, dict):
            return api_model(**{a: api_model.from_value(b) for a, b in value.items()})
        elif isinstance(value, list):
            return [api_model.from_value(v) for v in value]
        else:
            return value

    def __init__(self, **fields):
        self.members = []
        self.search = None
        for name, value in fields.items():
            setattr(self, name, value)

    def get_method(self, op, mm: api_model = None):
        if mm is None:
            return f"{self.name}_{op}"
        else:
            return f"{self.name}_{op}_{mm.name}"


api_models = {
    "Dataset": {
        "name": "dataset",
        "collection_name": "datasets",
        "search": "solr_search",
    },
    "Resource": {
        "name": "resource",
        "collection_name": "resources",
        "search": "resource_search",
    },
    "Organization": {
        "name": "organization",
        "collection_name": "organizations",
        "members": ["Dataset", "Workflow", "Tool", "Group", "User"],
    },
    "Group": {
        "name": "group",
        "collection_name": "groups",
        "members": ["Dataset", "Workflow", "Tool", "Group", "User"],
    },
    "Vocabulary": {
        "name": "vocabulary",
        "collection_name": "vocabularies",
    },
    "Tag": {
        "name": "tag",
        "collection_name": "tags",
    },
    "User": {
        "name": "user",
        "collection_name": "users",
    },
    "Process": {
        "name": "process",
        "collection_name": "processes",
        "members": ["Task"],
        "search": "solr_search",
    },
    "Task": {
        "name": "task",
        "collection_name": "tasks",
    },
    "Workflow": {
        "name": "workflow",
        "collection_name": "workflows",
        "search": "solr_search",
    },
    "Tool": {
        "name": "tool",
        "collection_name": "tools",
        "search": "solr_search",
    },
    "Policy": {
        "name": "policy",
        "collection_name": "policy",
    },
    "ImageRegistryToken": {
        "name": "image_registry_token",
        "collection_name": "image_registry_token",
    },
}
for m in api_models:
    api_models[m] = api_model.from_value(api_models[m])
for m in api_models:
    api_models[m].members = [api_models[mm] for mm in api_models[m].members]


OPERATIONS = [
    "create",
    "show",
    "update",
    "patch",
    "delete",
    "list",
    "fetch",
    "purge",
    "search",
]
MEMBER_OPERATIONS = ["add", "remove", "list_members"]
SEARCH_OPERATIONS = ["solr_search", "resource_search"]


class api_call_base(api_context):
    """Access the STELAR API using a client or a proxy.

    This is the base class for api_call, defining the generic methods
    for all types of entities.
    """

    def __init__(self, arg: APIContext):
        super().__init__(arg)

    def request(
        self, method: str, endpoint: str, params: dict = None, *, json=None, **kwargs
    ):
        if json is None:
            if kwargs:
                json = dict(kwargs)
        else:
            json = json | kwargs

        # This may raise requests exceptions
        resp = self.client.api_request(method, endpoint, params=params, json=json)
        jsout = resp.json()
        match jsout:
            case {"success": True, "result": result}:
                return result
            case {"success": False, "error": error}:
                match resp.status_code:
                    case 404:
                        raise EntityNotFound(
                            self.proxy_type, self.proxy_id, f"{method} {endpoint}"
                        )
                    case _:
                        raise ProxyOperationError(
                            self.proxy_type,
                            self.proxy_id,
                            f"{method} {endpoint}",
                            error,
                        )
            case _:
                raise RuntimeError(
                    "Unexpected response from the server",
                    method,
                    endpoint,
                    params,
                    json,
                    jsout,
                    resp.status_code,
                    resp,
                )

    def get_call(self, proxy_type, op, member_type=None):
        m = api_models[proxy_type.__name__]
        if member_type is None:
            call_name = m.get_method(op)
        else:
            mm = api_models[member_type.__name__]
            call_name = m.get_method(op, mm)
        return getattr(self, call_name)


# Populate the api_call class with the STELAR API endpoints
def generate_list(model: api_model):
    def gen_list(self, limit=None, offset=None):
        verb = "GET"
        endpoint = f"v2/{model.collection_name}"
        params = {"limit": limit, "offset": offset}
        return self.request(verb, endpoint, params)

    return gen_list


def generate_fetch(model: api_model):
    def gen_fetch(self, limit=None, offset=None):
        verb = "GET"
        endpoint = f"v2/{model.collection_name}.fetch"
        params = {"limit": limit, "offset": offset}
        return self.request(verb, endpoint, params)

    return gen_fetch


def generate_show(model: api_model):
    def gen_show(self, id):
        verb = "GET"
        endpoint = f"v2/{model.name}/{id}"
        return self.request(verb, endpoint)

    return gen_show


def generate_create(model: api_model):
    def gen_create(self, **kwargs):
        verb = "POST"
        endpoint = f"v2/{model.name}"
        return self.request(verb, endpoint, json=kwargs)

    return gen_create


def generate_update(model: api_model):
    def gen_update(self, id, **kwargs):
        verb = "PUT"
        endpoint = f"v2/{model.name}/{id}"
        return self.request(verb, endpoint, json=kwargs)

    return gen_update


def generate_patch(model: api_model):
    def gen_patch(self, id, **kwargs):
        verb = "PATCH"
        endpoint = f"v2/{model.name}/{id}"
        return self.request(verb, endpoint, json=kwargs)

    return gen_patch


def generate_delete(model: api_model):
    def gen_delete(self, id):
        verb = "DELETE"
        endpoint = f"v2/{model.name}/{id}"
        return self.request(verb, endpoint)

    return gen_delete


def generate_purge(model: api_model):
    def gen_purge(self, id):
        verb = "DELETE"
        endpoint = f"v2/{model.name}/{id}?purge=true"
        return self.request(verb, endpoint)

    return gen_purge


def generate_add(model: api_model, mm: api_model):
    def gen_add(self, id, member_id, capacity=None):
        verb = "POST"
        endpoint = f"v2/{model.name}/{id}/{mm.name}/{member_id}"
        return self.request(verb, endpoint, json={"capacity": capacity})

    return gen_add


def generate_remove(model: api_model, mm: api_model):
    def gen_remove(self, id, member_id):
        verb = "DELETE"
        endpoint = f"v2/{model.name}/{id}/{mm.name}/{member_id}"
        return self.request(verb, endpoint)

    return gen_remove


def generate_list_members(model: api_model, mm: api_model):
    def gen_list_members(self, id, capacity=None):
        verb = "GET"
        endpoint = f"v2/{model.name}/{id}/{mm.collection_name}"
        if capacity is not None:
            endpoint += f"?capacity={capacity}"
        return self.request(verb, endpoint)

    return gen_list_members


def generate_solr_search(model: api_model):
    def gen_search(self, query_spec):
        verb = "POST"
        endpoint = f"v2/search/{model.collection_name}"
        return self.request(verb, endpoint, json=query_spec)

    return gen_search


def generate_resource_search(model: api_model):
    def gen_search(self, query_spec):
        verb = "POST"
        endpoint = f"v2/search/{model.collection_name}"
        return self.request(verb, endpoint, json=query_spec)

    return gen_search


def generate_unimplemented(model: api_model, op, mm=None):
    def gen_unimplemented(self, *args, **kwargs):
        raise NotImplementedError(api_model.name, op, args, kwargs)

    return gen_unimplemented


# Instrumenting api_call_base with the generated methods.
# Where there is no specialized method defined,
# add the generated generic method to the api_call class.
for ptname in api_models:
    model = api_models[ptname]
    for op in OPERATIONS:
        call_name = model.get_method(op)
        match op:
            case "create":
                gcall = generate_create(model)
            case "show":
                gcall = generate_show(model)
            case "update":
                gcall = generate_update(model)
            case "patch":
                gcall = generate_patch(model)
            case "delete":
                gcall = generate_delete(model)
            case "list":
                gcall = generate_list(model)
            case "fetch":
                gcall = generate_fetch(model)
            case "purge":
                gcall = generate_purge(model)
            case "search":
                if model.search == "solr_search":
                    gcall = generate_solr_search(model)
                elif model.search == "resource_search":
                    gcall = generate_resource_search(model)
                else:
                    gcall = generate_unimplemented(model, op)

        gcall.__qualname__ = f"api_call_base.{call_name}"
        gcall.__name__ = call_name
        setattr(api_call_base, call_name, gcall)

    # Add the generated member methods to the api_call class
    for mm in model.members:
        for op in MEMBER_OPERATIONS:
            call_name = model.get_method(op, mm)

            match op:
                case "add":
                    gcall = generate_add(model, mm)
                case "remove":
                    gcall = generate_remove(model, mm)
                case "list_members":
                    gcall = generate_list_members(model, mm)

            gcall.__qualname__ = f"api_call_base.{call_name}"
            gcall.__name__ = call_name
            setattr(api_call_base, call_name, gcall)


class api_call(api_call_base):
    """Class that exposes the STELAR API for a given entity.

    `api_call(proxy).foo(...)`
    returns the 'result' of the STELAR API response on success,
    and raises a ProxyOperationError on failure.

    `api_call(client).foo(...)`
    does the same.
    """

    def __init__(self, arg: APIContext):
        super().__init__(arg)

    # def tag_list(self, vocabulary_id: str = None):
    #    raise NotImplementedError("tag_list")

    def user_fetch(self, limit: int = None, offset: int = None):
        users = self.request(
            "GET", "v1/users/", params={"limit": limit, "offset": offset}
        )
        return users

    def user_list(self, limit: int = None, offset: int = None):
        return [u["username"] for u in self.user_fetch()]

    def user_show(self, id: str):
        return self.request("GET", f"v1/users/{id}")

    def user_delete(self, id):
        return self.request("DELETE", f"v1/users/{id}")

    def user_create(self, **kwargs):
        return self.request("POST", "v1/users", json=kwargs)

    def user_update(self, id, **kwargs):
        raise NotImplementedError

    def user_patch(self, id, **kwargs):
        return self.request("PATCH", f"v1/users/{id}", json=kwargs)

    def user_purge(self, id):
        raise NotImplementedError

    def roles_fetch(self, limit: int = None, offset: int = None):
        """
        Fetch roles.

        Parameters
        ----------
        limit : int, optional
            The maximum number of roles to return.
        offset : int, optional
            The offset for pagination.

        Returns
        -------
        list
            A list of dictionaries containing role information.
        """
        roles = self.request("GET", "v1/users/roles")
        match (limit, offset):
            case (None, None):
                return roles
            case (None, _):
                return roles[offset:]
            case (_, None):
                return roles[:limit]
            case (_, _):
                return roles[offset : offset + limit]

    def user_add_role(self, user_id: str, role: str):
        """
        Add a role to a user.

        Parameters
        ----------
        user_id : str
            The ID of the user to whom the role will be added.
        role : str
            The role to be added to the user.
        Returns
        -------
        dict
            A dictionary containing the current user state.
        """
        return self.request("POST", f"v1/users/{user_id}/roles/{role}")

    def user_remove_role(self, user_id: str, role: str):
        """
        Remove a role from a user.

        Parameters
        ----------
        user_id : str
            The ID of the user from whom the role will be removed.
        role : str
            The role to be removed from the user.
        Returns
        -------
        dict
            A dictionary containing the current user state.
        """
        return self.request("DELETE", f"v1/users/{user_id}/roles/{role}")

    def user_add_roles(self, user_id: str, roles: list[str]):
        """
        Add multiple roles to a user.

        Parameters
        ----------
        user_id : str
            The ID of the user to whom the roles will be added.
        roles : list[str]
            A list of roles to be added to the user.

        Returns
        -------
        dict
            A dictionary containing the current user state.
        """
        return self.request("POST", f"v1/users/{user_id}/roles", json={"roles": roles})

    def user_set_roles(self, user_id: str, roles: list[str]):
        """
        Set roles for a user, replacing any existing roles.

        Parameters
        ----------
        user_id : str
            The ID of the user whose roles will be set.
        roles : list[str]
            A list of roles to be set for the user.

        Returns
        -------
        dict
            A dictionary containing the current user state.
        """
        return self.request("PATCH", f"v1/users/{user_id}/roles", json={"roles": roles})

    def entity_search(self, proxy_type: str, query_spec: dict):
        entity_type = api_models.collection_name
        return self.request("POST", f"v2/search/{entity_type}", json=query_spec)

    def dataset_export_zenodo(self, dataset_id: str) -> dict:
        """
        Export a dataset to Zenodo.

        Parameters
        ----------
        dataset_id : str
            The ID of the dataset to export.

        Returns
        -------
        dict
            A dictionary containing the export message, ready to be sent to zenodo.
        """
        return self.request("GET", f"v2/export/zenodo/{dataset_id}")

    #
    #
    # Handling tasks
    #
    #

    def task_job_input(self, task_id: str, signature: str) -> dict:
        """Get the input for a job in a task."""
        return self.request("GET", f"v2/task/{task_id}/{signature}/input")

    def task_post_job_output(
        self, task_id: str, signature: str, output_spec: dict
    ) -> dict:
        """Get the output for a job in a task."""
        return self.request(
            "POST", f"v2/task/{task_id}/{signature}/output", json=output_spec
        )

    def task_show_jobs(self, task_id: str):
        """Show the jobs associated with a task."""
        return self.request("GET", f"v2/task/{task_id}/jobs")

    def task_show_logs(self, task_id: str):
        """Show the logs associated with a task."""
        return self.request("GET", f"v2/task/{task_id}/logs")

    def task_signature(self, task_id: str) -> dict:
        """Get the signature of a task."""
        return self.request("GET", f"v2/task/{task_id}/signature")

    def task_list(self, limit: int = None, offset: int = None, state: str = None):
        """List tasks."""

        p = {"limit": limit, "offset": offset, "state": state}

        return self.request(
            "GET", "v2/tasks", params={k: v for k, v in p.items() if v is not None}
        )

    #
    # Policies
    #

    def policy_fetch(self, limit: int = None, offset: int = None):
        """
        Fetch policies.

        Parameters
        ----------
        limit : int, optional
            The maximum number of policies to return.
        offset : int, optional
            The offset for pagination.

        Returns
        -------
        list
            A list of dictionaries containing policy information.
        """
        policies = self.request("GET", "v1/auth/policy")["policies"]
        match (limit, offset):
            case (None, None):
                return policies
            case (None, _):
                return policies[offset:]
            case (_, None):
                return policies[:limit]
            case (_, _):
                return policies[offset : offset + limit]

    def policy_list(self, limit: int = None, offset: int = None):
        """
        List all policies.

        Parameters
        ----------
        limit : int, optional
            The maximum number of policies to return.
        offset : int, optional
            The offset for pagination.

        Returns
        -------
        list
            A list of dictionaries containing policy information (policy_uuid, policy_familiar_name).
        """
        return [e["policy_uuid"] for e in self.policy_fetch(limit, offset)]

    def policy_show(self, eid: str):
        """
        Show a specific policy.

        Parameters
        ----------
        policy_uuid : str
            The UUID of the policy to show.

        Returns
        -------
        dict
            A dictionary containing the policy information.
        """
        return self.request("GET", f"v1/auth/policy/{eid}")

    def policy_create(self, policy_yaml: str | bytes):
        """
        Create a new policy.

        Parameters
        ----------
        prolicy_data: str | bytes

        Returns
        -------
        dict
            A dictionary containing the created policy information.
        """

        # We need to use the client.request method in order to send
        # the policy data as a string or bytes.
        headers = {
            "Content-Type": "application/x-yaml",
        }
        response = self.client.request(
            "POST", "api/v1/auth/policy", headers=headers, data=policy_yaml
        )
        if response.status_code in range(200, 300):
            return response.json()["result"]
        else:
            raise RuntimeError(
                "Unexpected response trying to create policy",
                policy_yaml,
                response.status_code,
                response.json(),
            )

    def policy_spec(self, policy_uuid: str) -> bytes:
        """
        Get the specification of a policy.

        Parameters
        ----------
        policy_uuid : str
            The UUID of the policy to get the specification for.

        Returns
        -------
        dict
            A dictionary containing the policy specification.
        """
        response = self.client.GET("v1/auth/policy/representation", policy_uuid)
        if response.status_code in range(200, 300):
            return response.content
        else:
            raise RuntimeError(
                "Unexpected response trying to get policy spec",
                policy_uuid,
                response.status_code,
                response,
            )

    ##
    #
    # Image Registry Tokens
    #

    def image_registry_token_list(self, limit: int = None, offset: int = None):
        """
        List image registry tokens.

        Parameters
        ----------
        limit : int, optional
            The maximum number of tokens to return.
        offset : int, optional
            The offset for pagination.

        Returns
        """
        result = self.request("GET", "v2/registry/credentials")
        tokens = list(result.keys())
        match (limit, offset):
            case (None, None):
                return tokens
            case (None, _):
                return tokens[offset:]
            case (_, None):
                return tokens[:limit]
            case (_, _):
                return tokens[offset : offset + limit]

    def image_registry_token_create(self, title: str, expiration: str | None = None):
        """
        Create a new image registry token.
        Parameters
        ----------
        title : str
            The title of the token.
        expiration : str | None, optional
            The expiration date of the token in ISO format. Defaults to None.

        Returns
        -------
        dict
            A dictionary containing the created token information.
        """
        json = {"title": title}
        # TODO: Fix the stelarapi to accept expiration as a string

        return self.request("POST", "v2/registry/credentials", json=json)

    def image_registry_token_show(self, id: str):
        """
        Show an image registry token.

        Parameters
        ----------
        uuid : str
            The UUID of the token to show.

        Returns
        -------
        dict
            A dictionary containing the token information.
        """
        return self.request("GET", f"v2/registry/credentials/{id}")

    def image_registry_token_delete(self, uuid: str):
        """
        Delete an image registry token.

        Parameters
        ----------
        uuid : str
            The UUID of the token to delete.

        Returns
        -------
        dict
            A dictionary containing the deletion result.
        """
        return self.request("DELETE", f"v2/registry/credentials/{uuid}")

    #
    #
    # Relationships
    #
    #

    def relationships_fetch(
        self,
        subject_id: str,
        rel: Optional[str] = None,
        object_id: Optional[str] = None,
        /,
    ):
        """
        Show relationships for a subject.

        """
        p = [str(subject_id)]
        if rel is not None:
            p.append(str(rel))
            if object_id is not None:
                p.append(str(object_id))

        endpoint = f"v2/relationships/{'/'.join(p)}"
        return self.request("GET", endpoint)

    def relationship_show(self, subject_id: str, rel: str, object_id: str):
        """
        Show a specific relationship.

        Parameters
        ----------
        subject_id : str
            The ID of the subject of the relationship.
        rel : str
            The type of the relationship.
        object_id : str
            The ID of the object of the relationship.

        Returns
        -------
        dict
            A dictionary containing the relationship information.
        """
        r = self.relationships_fetch(subject_id, rel, object_id)
        if not r:
            raise EntityNotFound(
                "Relationship not found",
                "relationships",
                (subject_id, rel, object_id),
                "fetch",
            )
        return r[0]

    def relationship_create(
        self,
        subject_id: str,
        rel: str,
        object_id: str,
        comment: Optional[str] = None,
    ):
        """
        Create a new relationship.

        Parameters
        ----------
        subject_id : str
            The ID of the subject of the relationship.
        rel : str
            The type of the relationship.
        object_id : str
            The ID of the object of the relationship.
        comment : str, optional
            A comment for the relationship.

        Returns
        -------
        dict
            A dictionary containing the created relationship information.
        """
        return self.request(
            "POST",
            f"v2/relationship/{subject_id}/{rel}/{object_id}",
            json=dict(comment=comment),
        )

    def relationship_update(
        self,
        subject_id: str,
        rel: str,
        object_id: str,
        comment: Optional[str] = None,
    ):
        """
        Update the comment of a relationship.

        Parameters
        ----------
        subject_id : str
            The ID of the subject of the relationship.
        rel : str
            The type of the relationship.
        object_id : str
            The ID of the object of the relationship.
        comment : str, optional
            A comment for the relationship.

        Returns
        -------
        dict
            A dictionary containing the created relationship information.
        """
        return self.request(
            "PUT",
            f"v2/relationship/{subject_id}/{rel}/{object_id}",
            json=dict(comment=comment),
        )

    def relationship_delete(self, subject_id: str, rel: str, object_id: str):
        """
        Delete a relationship.

        Parameters
        ----------
        subject_id : str
            The ID of the subject of the relationship.
        rel : str
            The type of the relationship.
        object_id : str
            The ID of the object of the relationship.
        """
        self.request("DELETE", f"v2/relationship/{subject_id}/{rel}/{object_id}")



"""
    Facility for calling the STELAR API.

    The functions in this file will be changed (together with this documentation)
    as the API implementation progresses.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Type, TypeVar
from uuid import UUID

# from .api_call import api_call_DC as api_call
from .api_call import api_call
from .proxy import (
    EntityNotFound,
    ErrorState,
    Id,
    Proxy,
    ProxyCursor,
    ProxyList,
    ProxyState,
    ProxySynclist,
    Registry,
)

if TYPE_CHECKING:
    from .client import Client
ProxyClass = TypeVar("ProxyClass", bound=Proxy)


def generic_proxy_sync(proxy: Proxy, entity, update_method="patch"):
    """Perform proxy_sync(entity) using the api_call class.

    Args:
        proxy (Proxy): the proxy to sync
        entity (Entity|None): the entity to sync from
        update_method: one of 'update' or 'patch'
    """
    if update_method not in ("update", "patch"):
        raise ValueError("Update method must be either 'patch' or 'update'")
    proxy_type = type(proxy)
    ac = api_call(proxy)

    if proxy.proxy_state is ProxyState.ERROR:
        raise ErrorState(proxy)

    if proxy.proxy_id == UUID(int=0):
        # Create the entity !
        entity_properties = proxy.proxy_to_entity()
        # Populate sync list
        psl = ProxySynclist()
        psl.on_create_proxy(proxy)

        # Create the entity
        create = ac.get_call(type(proxy), "create")
        entity = create(**entity_properties)

        # We add this proxy to its registry
        proxy.proxy_registry.register_proxy_for_entity(proxy, entity)
        proxy.proxy_from_entity(entity)
        proxy.proxy_changed = None

        # Sync around
        psl.sync()

    else:
        # The entity exists, make a normal sync

        if proxy.proxy_changed is not None:
            # A proper update is needed
            if update_method == "patch":
                updates = proxy.proxy_to_entity(proxy.proxy_changed)
                update_call = ac.get_call(proxy_type, "patch")
            elif update_method == "update":
                updates = proxy.proxy_to_entity()
                update_call = ac.get_call(proxy_type, "update")

            try:
                entity = update_call(id=str(proxy.proxy_id), **updates)
            except EntityNotFound:
                proxy.proxy_is_purged()
                raise  # We have updates that are lost
            proxy.proxy_changed = None

        if entity is None:
            show = ac.get_call(proxy_type, "show")
            try:
                entity = show(id=str(proxy.proxy_id))
            except EntityNotFound:
                proxy.proxy_is_purged()
                return  # Not an error!

        proxy.proxy_from_entity(entity)


def generic_create(
    client: Client, proxy_type: Type[ProxyClass], **properties
) -> ProxyClass:
    entity_properties = proxy_type.new_entity(client, **properties)

    # Populate the sync list
    psl = ProxySynclist()
    psl.on_create(proxy_type, properties)

    # Create the entity
    ac = api_call(client)
    create = ac.get_call(proxy_type, "create")
    new_entity = create(**entity_properties)
    registry = client.registry_for(proxy_type)
    new_proxy = registry.fetch_proxy_for_entity(new_entity)

    # Sync around
    psl.sync()
    return new_proxy


def generic_get(
    client: Client, proxy_type: Type[ProxyClass], name_or_id, default=None
) -> ProxyClass:
    ac = api_call(client)
    show = ac.get_call(proxy_type, "show")
    try:
        entity = show(str(name_or_id))
    except EntityNotFound:
        return default
    return client.registry_for(proxy_type).fetch_proxy_for_entity(entity)


def generic_fetch_list(
    client: Client, proxy_type: Type[ProxyClass], *, limit: int, offset: int, **kwargs
) -> list[str]:
    ac = api_call(client)
    _list = ac.get_call(proxy_type, "list")
    result = _list(limit=limit, offset=offset, **kwargs)

    return result


def generic_fetch_old(
    client: Client, proxy_type: Type[ProxyClass], *, limit: int, offset: int, **kwargs
) -> Iterator[ProxyClass]:
    ac = api_call(client)
    _list = ac.get_call(proxy_type, "list", **kwargs)
    result = _list(limit=limit, offset=offset)

    registry = client.registry_for(proxy_type)

    # Check it the list already contains entities
    if result and isinstance(result[0], dict):
        for entity in result:
            yield registry.fetch_proxy_for_entity(entity)
    else:
        show = ac.get_call(proxy_type, "show")
        for name in result:
            entity = show(id=name)
            yield registry.fetch_proxy_for_entity(entity)


def generic_fetch(
    client: Client, proxy_type: Type[ProxyClass], *, limit: int, offset: int, **kwargs
) -> Iterator[ProxyClass]:
    ac = api_call(client)
    _fetch = ac.get_call(proxy_type, "fetch", **kwargs)
    result = _fetch(limit=limit, offset=offset)

    registry = client.registry_for(proxy_type)

    # Check it the list already contains entities
    for entity in result:
        yield registry.fetch_proxy_for_entity(entity)


def generic_delete(proxy: Proxy, purge=False):
    psl = ProxySynclist()
    psl.on_delete(proxy)

    ac = api_call(proxy)

    delete = ac.get_call(type(proxy), "purge" if purge else "delete")
    delete(id=str(proxy.proxy_id))
    psl.sync()


#
# The generic base classes for proxy and cursor
#


class GenericProxy(Proxy, entity=False):
    id = Id()

    def delete(self, purge=False):
        """
        Delete the entity proxied by this proxy.

        If the entity is purged then the deletion is permanent.
        After the purge deletion, the proxy is left in the ERROR state.

        Otherwise, the deletion removes the entity from the active
        ones, but the entity can still be accessed. Searches will
        not return the entity.

        TODO: Implement entity reclamation for non-purged entities!
        """
        generic_delete(self, purge=purge)
        super().delete(purge=purge)

    def proxy_sync(self, entity=None):
        """Sync the proxy with the entity in the API."""
        return generic_proxy_sync(self, entity)


class GenericProxyList(ProxyList):
    def __init__(self, eid_list, client, proxy_type):
        """A proxy list using a list of ids or names.

        The resolving is done using 'generic_get()'
        """

        super().__init__(client, proxy_type)
        self._data = eid_list

    @property
    def coll(self):
        return self._data

    def resolve_proxy(self, item):
        return generic_get(self.client, self.proxy_type, item)


class GenericCursor(ProxyCursor[ProxyClass], Registry[ProxyClass]):
    def create(self, **prop) -> ProxyClass:
        return self.proxy_type.new(self.client, **prop)

    def get(
        self, name_or_id: str | UUID, default: ProxyClass | None = None
    ) -> ProxyClass:
        return generic_get(self.client, self.proxy_type, name_or_id, default=default)

    def fetch_list(self, *, limit: int, offset: int) -> list[str]:
        return generic_fetch_list(
            self.client, self.proxy_type, limit=limit, offset=offset
        )

    def fetch(self, *, limit: int, offset: int) -> Iterator[ProxyClass]:
        yield from generic_fetch(
            self.client, self.proxy_type, limit=limit, offset=offset
        )

    def __getitem__(self, item):
        v = super().__getitem__(item)
        if isinstance(v, list):
            return GenericProxyList(v, self.client, self.proxy_type)
        else:
            return v

    def _ipython_key_completions_(self):
        try:
            kc = self.fetch_list(limit=100, offset=0)
        except Exception:
            kc = []
        return kc



from datetime import datetime, timedelta
from uuid import UUID, uuid4

import pytest

from stelar.client.proxy import *
from stelar.client.proxy.fieldvalidation import FieldValidator


def test_basic_checks():
    val = FieldValidator()
    assert val.validate(None) is None
    assert val.validate(2) == 2


def test_nullable():
    nv = AnyField(nullable=True)
    nnv = AnyField(nullable=False)

    assert nv.validate(None) is None
    with pytest.raises(ValueError):
        nnv.validate(None)


def test_minimum_value():
    v = AnyField(minimum_value=10)
    assert v.validate(None) is None
    assert v.validate(20) == 20
    assert v.validate(10) == 10
    with pytest.raises(ValueError):
        v.validate(9)
    with pytest.raises(ValueError):
        v.validate(0)


def test_maximum_value():
    v = AnyField(maximum_value=10)
    assert v.validate(None) is None
    assert v.validate(-20) == -20
    assert v.validate(10) == 10
    with pytest.raises(ValueError):
        v.validate(19)
    with pytest.raises(ValueError):
        v.validate(11)


def test_max_min_length():
    v = StrField(minimum_len=3, maximum_len=5)

    assert v.prioritized_checks == [
        (v.check_null, -1),
        (v.to_ftype, 5),
        (v.check_length, 20),
    ]

    assert v.validate("abc") == "abc"
    assert v.validate("abcde") == "abcde"

    for s in ["", "a", "sf", "123456", "adlasdjasdasdlas"]:
        with pytest.raises(ValueError):
            v.validate(s)


def test_str_field():
    v = StrField()
    assert v.validate(10) == "10"
    assert v.validate(True) == "True"
    assert v.validate("10 312") == "10 312"
    assert v.validate("") == ""


def test_int_field():
    v = IntField()

    assert v.validate(-1231) == -1231
    assert v.validate(1.4) == 1
    assert v.validate(False) == 0
    assert v.validate("14") == 14

    with pytest.raises(ValueError):
        v.validate("aaa")


def test_bool_field():
    v = BoolField()

    assert v.validate(True) is True
    assert v.validate(False) is False
    assert v.validate(-1231) is True
    assert v.validate(1.4) is True
    assert v.validate(0) is False
    assert v.validate(0.0) is False
    assert v.validate([]) is False
    assert v.validate("14") is True
    assert v.validate("") is False


def test_no_checks():
    v = FieldValidator(nullable=None)
    assert v.prioritized_checks == []
    assert v.checks == []


def test_strict():
    v = AnyField(nullable=None, strict=True)
    assert v.prioritized_checks == []
    assert v.checks == []

    for s in [None, 1, "aa", ...]:
        with pytest.raises(ValueError):
            v.validate(s)

    assert v.convert_to_entity(10) == 10
    assert v.convert_to_proxy(10) == 10


def test_date_field():
    v = DateField(nullable=False, minimum_value=datetime(2020, 1, 1))

    n = datetime.now()
    assert v.validate(n) == n
    tendays = timedelta(days=10)
    assert v.validate(n - tendays) == n - tendays

    with pytest.raises(ValueError):
        v.validate(datetime(2019, 12, 31, 23, 59))
    with pytest.raises(ValueError):
        v.validate(None)

    assert v.validate(n.isoformat()) == n
    assert v.validate("2022-11-11") == datetime.fromisoformat("2022-11-11")
    with pytest.raises(ValueError):
        v.validate("2022/11/11")
    with pytest.raises(ValueError):
        v.validate(12313)

    assert v.convert_to_entity(n) == n.isoformat()
    assert v.convert_to_proxy(n.isoformat()) == n


def test_uuid_field():
    v = UUIDField()

    u = uuid4()
    assert v.validate(str(u)) == u
    assert v.validate(u) == u
    with pytest.raises(ValueError):
        v.validate(1192929292)

    assert v.convert_to_entity(u) == str(u)
    assert v.convert_to_proxy(str(u)) == u


def test_state_field():
    v = StateField()

    assert v.validate("active")
    assert v.validate("deleted")

    for val in ("", "ACTIVE", 4, None):
        with pytest.raises(ValueError):
            assert v.validate(val)



import re
import time
from typing import Type, TypeVar
from urllib.parse import urlencode, urljoin

import requests

from .proxy import Proxy, ProxyCursor, RegistryCatalog

ProxyClass = TypeVar("ProxyClass", bound="Proxy")


class KLMSInfo:
    def __init__(self, info: dict):
        self.__dict__.update(info)

    def __repr__(self):
        return f"<KLMSInfo {self.__dict__}>"

    def __str__(self):
        return str(self.__dict__)

    def _repr_pretty_(self, p, cycle: bool):
        for k, v in self.__dict__.items():
            p.text(f"{k}: {v}")
            p.break_()


class BaseAPI(RegistryCatalog):
    """Base class for all parts of the client API.

    Its main responsibility is to support API calls to the
    STELAR server. It also contains logic to manage the proxies.
    """

    def __init__(self, base_url, token_json, tls_verify=True):
        super().__init__()
        self._base_url = base_url
        self._api_url = base_url + "/api/"
        self._tls_verify = tls_verify
        self._http_session = requests.Session()
        self.reset_tokens(token_json)

        # Initialize the default organization
        self._info = self._get_api_info()

    def _get_api_info(self):
        help = self.request("GET", "help").json()["result"]
        info_attr = {
            k: v for k, v in help.items() if re.match(r"[A-Za-z][A-Za-z0-9_]*", k)
        }
        from importlib.metadata import version

        info_attr["client_version"] = version("stelar.client")
        return KLMSInfo(info_attr)

    @property
    def klms_info(self):
        return self._info

    @property
    def api_url(self):
        """Return the base URL to the STELAR API"""
        return self._api_url

    @property
    def token(self):
        """Getter for the token property."""
        return self._token

    def reset_tokens(self, token_json):
        """
        Reset the primary and refresh tokens.

        Args:
            token_json: The token response for authentication
        """
        token = token_json.get("token")
        refresh_token = token_json.get("refresh_token")
        expires_in = token_json.get("expires_in")
        refresh_expires_in = token_json.get("refresh_expires_in")

        self._token_json = token_json
        self._token = token
        self._refresh_token = refresh_token
        self._http_session.headers.update({"Authorization": f"Bearer {self._token}"})

        self._expires_in = expires_in
        self._refresh_expires_in = refresh_expires_in
        self._token_reset_time = time.time()
        self._expiration_time = self._token_reset_time + self._expires_in * 0.9
        self._refresh_expiration_time = (
            self._token_reset_time + self._refresh_expires_in * 0.9
        )

    def token_expired(self):
        """
        Check if the token has expired.

        Returns:
            bool: True if the token has expired, False otherwise.
        """
        return time.time() > self._expiration_time

    def cursor_for(
        self, proxy_type: Type[ProxyClass] | str, **kwargs
    ) -> ProxyCursor[ProxyClass]:
        """Get a cursor for a given proxy type.

        Args:
            proxy_type: The type of the proxy. This can be provided as a string
                (the name of the proxy) or as the class itself.
            **kwargs: Additional arguments to pass to the cursor.

        Returns:
            ProxyCursor: The cursor for the proxy type.

        Example:
            client.cursor_for("Dataset")
            client.cursor_for(Dataset)
        """
        return self.registry_for(proxy_type)

    def request(
        self, method, endpoint, params=None, data=None, headers=None, json=None
    ):
        """
        Sends a request to the STELAR API.

        The main difference of this method with respect to `api_request` is that the
        endpoint is relative to the base URL, and it can include headers, as well as
        non-JSON data (e.g., form data).

        For example:
            response = c.request("GET", "api/v2/datasets")
            response = c.request("POST", "v2/dataset", json={"name": "my_dataset"})

        Args:
            method (str): The HTTP method ('GET', 'POST', 'PUT', 'DELETE').
            endpoint (str): The API endpoint (relative to `api_url`). Can include query parameters.
            params (dict, optional): URL query parameters.
            data (dict, optional): Form data to be sent in the body.
            headers (dict, optional): Additional request headers.
            json (dict, optional): JSON data to be sent in the body.

        Returns:
            requests.Response: The response object from the API.
        """
        # Validate data/json and handle accordingly
        if method.upper() == "GET":
            # GET requests should not have a body (data or json)
            if data or json:
                raise ValueError("GET requests cannot include body data.")
        else:
            # POST, PUT, DELETE, etc., should use either data or json but not both
            if data and json:
                raise ValueError("Specify either 'data' or 'json', not both.")

        # Handle query parameters in the endpoint or passed as 'params'
        if "?" in endpoint and params:
            raise ValueError(
                "Specify query parameters either in the endpoint or in 'params', not both."
            )

        # Combine base_url with the endpoint
        endpoint = endpoint.lstrip("/")
        url = urljoin(self._base_url + "/", endpoint)
        # If the URL does not contain a query, add parameters from 'params'
        if params:
            url = f"{url}?{urlencode(params)}"

        # Prepare headers, defaulting to Authorization if token is present and Content-Type
        default_headers = {
            "Content-Type": "application/json",
        }

        # if self._token:
        #    default_headers["Authorization"] = f"Bearer {self._token}"

        if headers:
            default_headers.update(headers)

        if self.token_expired():
            self.reauthenticate()

        turn = 0
        while turn < 2:
            # Make the request using the provided method, url, params, data, json, and headers
            # response = requests.request(
            response = self._http_session.request(
                method=method,
                url=url,
                params=None,  # params are already incorporated into the URL
                data=data,  # if provided, this will be form data
                json=json,  # if provided, this will be JSON payload
                headers=default_headers,
                verify=self._tls_verify,
            )
            if response.status_code == 401 and turn == 0:
                # Refresh the token and try again
                self.reauthenticate()
                # default_headers["Authorization"] = f"Bearer {self._token}"
                turn += 1
            else:
                break

        # Return the response
        return response

    def api_request(self, method, endpoint, *, params=None, json=None):
        """Do an actual API call

        Examples:
            response = api_request("GET", "v2/datasets")
            response = api_request("POST", "v2/dataset", json={"name": "my_dataset"})

        Args:
            method (str): The HTTP method ('GET', 'POST', 'PUT', 'PATCH', 'DELETE').
            endpoint (str): The API endpoint. It should be given as a relative path.
            params (dict, optional): URL query parameters.
            json (dict, optional): JSON data to be sent in the body.
        """

        url = urljoin(self._api_url, endpoint)
        # headers = {"Authorization": f"Bearer {self._token}"}
        if params is not None and "json" in params:
            json = params["json"]

        if self.token_expired():
            self.reauthenticate()

        twice = 0
        while twice < 2:
            response = self._http_session.request(
                method,
                url,
                params=params,
                json=json,
                verify=self._tls_verify,
            )
            if response.status_code == 401 and twice == 0:
                self.reauthenticate()
                twice += 1
            else:
                break

        return response

    def GET(self, *endp, **params):
        """Send a GET request to the API.

        Examples:
            response = c.GET("v2/datasets")
            response = c.GET("v2/dataset", 'my_dataset')

        Args:
            *endp: Path components to the API endpoint.
            **params: The query parameters to send.

        Returns:
            requests.Response: The response object from the API.
        """
        endpoint = "/".join(str(pc) for pc in endp)
        return self.api_request("GET", endpoint, params=params)

    def POST(self, *endp, params={}, **json):
        """Send a POST request to the API.

        Example:
            response = c.POST("v2/dataset", name="my_dataset")

        Args:
            *endp: Path components to the API endpoint.
            params: (keyword-only) The query parameters to send.
            **json: The JSON data to send.

        Returns:
            requests.Response: The response object from the API.
        """
        endpoint = "/".join(str(pc) for pc in endp)
        return self.api_request("POST", endpoint, params=params, json=json)

    def PUT(self, *endp, params={}, **json):
        """Send a PUT request to the API.

        Example:
            response = c.PUT("v2/tag", "my_tag", display_name="My Tag", vocabulary_id='my_vocabulary')

        Args:
            *endp: Path components to the API endpoint.
            params: (keyword-only) The query parameters to send.
            **json: The JSON data to send.

        Returns:
            requests.Response: The response object from the API.
        """
        endpoint = "/".join(str(pc) for pc in endp)
        return self.api_request("PUT", endpoint, params=params, json=json)

    def PATCH(self, *endp, params={}, **json):
        """Send a PATCH request to the API.

        Example:
            response = c.PATCH("v2/dataset", "my_dataset", author="John Doe")

        Args:
            *endp: Path components to the API endpoint.
            params: (keyword-only) The query parameters to send.
            **json: The JSON data to send.
        Returns:
            requests.Response: The response object from the API.
        """
        endpoint = "/".join(str(pc) for pc in endp)
        return self.api_request("PATCH", endpoint, params=params, json=json)

    def DELETE(self, *endp, **params):
        """Send a DELETE request to the API.

        Example:
            response = c.DELETE("v2/dataset", "my_dataset")

        Args:
            *endp: Path components to the API endpoint.
            **params: The query parameters to send.
        Returns:
            requests.Response: The response object from the API.
        """
        endpoint = "/".join(str(pc) for pc in endp)
        return self.api_request("DELETE", endpoint, params=params)

    @property
    def api(self):
        from .api_call import api_call

        return api_call(self)

    def listapi(self, show_verbs=True):
        """Return a list of all available API endpoints."""
        specs = self.GET("../specs").json()
        paths = specs["paths"]

        L = []
        for path, methods in paths.items():
            if show_verbs:
                for method in methods:
                    L.append(f"{method.upper():<8} {path}")
            else:
                L.append(path)

        return L



from __future__ import annotations

from typing import TYPE_CHECKING

from .decl import tag_split
from .fieldvalidation import AnyField
from .property import Property
from .proxy import Proxy

if TYPE_CHECKING:
    from .registry import RegistryCatalog
    from .typing import TagDictList, TagSpecList


class VocabularyIndex:
    """This is a component to be included as part of a registry catalog.
    It is responsible for indexing the tag vocabularies of a STELAR API,
    translating vocabulary names to IDs and back. As vocabularies do not
    change often at all, this index should only be included once in the lifetime
    of a client.
    """

    def __init__(self, catalog: RegistryCatalog, *args, **kwargs):
        self.catalog = catalog
        self.dirty = True
        self.refresh_count = 0
        self._name_to_id = {}
        self._id_to_name = {}
        self._name_to_tags = {}
        self._id_to_tags = {}

    @property
    def name_to_tags(self):
        if self.dirty:
            self.refresh_tag_vocabularies()
        return self._name_to_tags

    @property
    def id_to_tags(self):
        if self.dirty:
            self.refresh_tag_vocabularies()
        return self._id_to_tags

    @property
    def name_to_id(self):
        if self.dirty:
            self.refresh_tag_vocabularies()
        return self._name_to_id

    @property
    def id_to_name(self):
        if self.dirty:
            self.refresh_tag_vocabularies()
        return self._id_to_name

    def refresh_tag_vocabularies(self):
        """The registry catalog"""
        self._id_to_name.clear()
        self._name_to_id.clear()

        for voc in self.catalog.fetch_active_vocabularies():
            vid = voc["id"]
            vname = voc["name"]
            tags = {tag["name"]: tag for tag in voc["tags"]}
            self._id_to_name[vid] = vname
            self._name_to_id[vname] = vid
            self._name_to_tags[vname] = tags
            self._id_to_tags[vid] = tags
        self.dirty = False
        self.refresh_count += 1

    def validate_tagspec(self, tagspec: str) -> bool:
        """Check if a string is formatted correctly as a tagspec"""
        voc, tag = tag_split(tagspec)
        return voc is None or tag in self.name_to_tags.get(voc, ())

    def tagspec_to_id(self, tagspec: str) -> str:
        """Convert a tagspec to a tag ID"""
        voc, tagname = tag_split(tagspec)
        if voc is None:
            return tagname
        return self.name_to_tags[voc][tagname]["id"]


class TagListField(AnyField):
    def __init__(self):
        super().__init__(self, nullable=False, default=())
        self.add_check(self.to_taglist, 5)

    def to_taglist(self, tagl, *, vocindex, **kwargs):
        if isinstance(tagl, str):
            raise ValueError("An iterable of strings was expected, got a single string")
        taglist = tuple(tagl)
        if not all(isinstance(v, str) for v in taglist):
            raise ValueError(
                "Unexpected tagspec", [v for v in taglist if not isinstance(v, str)]
            )
        badts = [ts for ts in taglist if not vocindex.validate_tagspec(ts)]
        if badts:
            raise ValueError("Tag list contains non-valid tagspec(s)", badts)
        return taglist, True

    def convert_to_entity(
        self, value: TagSpecList, *, vocindex: VocabularyIndex, **kwargs
    ) -> TagDictList:
        return list(value)

    def convert_to_proxy(
        self, value: TagDictList, *, vocindex: VocabularyIndex, **kwargs
    ) -> TagSpecList:
        return tuple(value)


class TagList(Property):
    def __init__(self, doc="The tag list", short=True, **kwargs):
        super().__init__(
            validator=TagListField,
            updatable=True,
            optional=False,
            doc=doc,
            short=short,
            **kwargs,
        )

    def validate(self, proxy, value):
        vocindex = proxy.proxy_registry.catalog.vocabulary_index
        return self.validator.validate(value, vocindex=vocindex)

    def convert_entity_to_proxy(self, proxy, entity, **kwargs):
        vocindex = proxy.proxy_registry.catalog.vocabulary_index
        return super().convert_entity_to_proxy(
            proxy, entity, vocindex=vocindex, **kwargs
        )

    def convert_proxy_to_entity(self, proxy, entity, **kwargs):
        vocindex = proxy.proxy_registry.catalog.vocabulary_index
        return super().convert_proxy_to_entity(
            proxy, entity, vocindex=vocindex, **kwargs
        )

    def convert_to_create(
        self, proxy_type, create_props, entity_props, *, catalog, **kwargs
    ):
        vocindex = catalog.vocabulary_index
        return super().convert_to_create(
            proxy_type, create_props, entity_props, vocindex=vocindex, **kwargs
        )


class TaggableProxy(Proxy, entity=False):
    """A virtual base class for all proxies to entities that are taggable."""

    tags = TagList()

    def add_tags(self, *tags: str):
        """Add multiple tags to this entity.

        Args:
            *tags (str): The tags to add, in tagspec format.
        """
        self.tags += tags

    def remove_tags(self, *tags: str):
        """Remove multiple tags from this entity.

        Args:
            *tags (str): The tags to remove, in tagspec format.
        """
        current_tags = set(self.tags)
        tags_to_remove = set(tags)
        self.tags = tuple(current_tags - tags_to_remove)



import pytest

from stelar.client import Client
from stelar.client.base import KLMSInfo

#
#  A fixture with a sample context file
#


@pytest.fixture(scope="session")
def config_file(tmp_path_factory):
    ctxfile = tmp_path_factory.mktemp("data") / ".stelar"
    ctx = """
; Test context
[default]
base_url= http://klms.example.org/
username= joe
password= joesecret

; Another test context
[biglake]
base_url= http://big.example.org/
username= joe
password= joesecret
"""
    with open(ctxfile, "w") as f:
        f.write(ctx)
    return ctxfile


def test_normalize_urls(mocker):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://foo.bar.com"}),
    )

    cli = Client(base_url="http://user:pass@foo.bar.com")
    assert cli._base_url == "http://foo.bar.com/stelar"
    assert cli._api_url == "http://foo.bar.com/stelar/api/"

    cli = Client(base_url="https://user:pass@foo.bar.com/")
    assert cli._base_url == "https://foo.bar.com/stelar"
    assert cli._api_url == "https://foo.bar.com/stelar/api/"

    cli = Client(base_url="https://user:pass@foo.bar.com/stelar")
    assert cli._base_url == "https://foo.bar.com/stelar"
    assert cli._api_url == "https://foo.bar.com/stelar/api/"


def test_context_default(mocker, config_file):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://minio.example.org"}),
    )

    c = Client(config_file=config_file)
    assert c._context is "default"
    assert c._base_url == "http://klms.example.org/stelar"
    Client.authenticate.assert_called_once_with(
        "http://klms.example.org/stelar",
        username="joe",
        password="joesecret",
        tls_verify=True,
    )


def test_context_given(mocker, config_file):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://minio.example.org"}),
    )

    c = Client(context="default", config_file=config_file)

    assert c._context == "default"
    assert c._base_url == "http://klms.example.org/stelar"
    Client.authenticate.assert_called_once_with(
        "http://klms.example.org/stelar",
        username="joe",
        password="joesecret",
        tls_verify=True,
    )


def test_context_noverify(mocker, config_file):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://minio.example.org"}),
    )

    c = Client(context="default", config_file=config_file, tls_verify=False)

    assert c._context == "default"
    assert c._base_url == "http://klms.example.org/stelar"
    Client.authenticate.assert_called_once_with(
        "http://klms.example.org/stelar",
        username="joe",
        password="joesecret",
        tls_verify=False,
    )
    assert c._tls_verify is False


def test_user_pass_in_url(mocker):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://minio.example.org"}),
    )

    c = Client(base_url="https://joe:joesecret@foo.bar.com")

    assert c._base_url == "https://foo.bar.com/stelar"
    assert c._token == "token"
    assert c._refresh_token == "refresh"
    c.authenticate.assert_called_once_with(
        "https://foo.bar.com/stelar",
        username="joe",
        password="joesecret",
        tls_verify=True,
    )


def test_no_user_pass(mocker):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://minio.example.org"}),
    )

    c = Client(base_url="https://foo.bar.com")

    assert c._base_url == "https://foo.bar.com/stelar"
    assert c._token == "token"
    assert c._refresh_token == "refresh"
    c.authenticate.assert_called_once_with(
        "https://foo.bar.com/stelar", username="", password="", tls_verify=True
    )


def test_user_pass_given(mocker):
    mocker.patch(
        "stelar.client.Client.authenticate",
        return_value={
            "token": "token",
            "refresh_token": "refresh",
            "expires_in": 100,
            "refresh_expires_in": 100,
        },
    )
    mocker.patch(
        "stelar.client.base.BaseAPI._get_api_info",
        return_value=KLMSInfo({"s3_api": "http://minio.example.org"}),
    )

    c = Client(base_url="https://foo.bar.com", username="joe", password="joesecret")

    assert c._base_url == "https://foo.bar.com/stelar"
    assert c._token == "token"
    assert c._refresh_token == "refresh"
    c.authenticate.assert_called_once_with(
        "https://foo.bar.com/stelar",
        username="joe",
        password="joesecret",
        tls_verify=True,
    )


@pytest.mark.skip
def test_fetch(testcli):
    dsets = testcli.catalog.get_datasets()
    assert len(dsets) > 0
    dset = dsets[0]
    print(dset.id)
    print(dset.name)
    print(dset.title)
    print(dset.notes)



"""
    Dummy conftest.py for stelar.client.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""


import pytest

from stelar.client import Client


@pytest.fixture(scope="session")
def testcontext() -> str:
    return "local"


def client_cleanup(cli):
    # Make sure that all datasets whose name
    # starts with "test_" are purged
    for m in cli.DC.package_autocomplete(q="test_", limit=1000)["result"]:
        if m["match_field"] == "name":
            cli.DC.dataset_purge(id=m["name"])


@pytest.fixture()
def testcli(testcontext) -> Client:
    cli = Client(testcontext)

    client_cleanup(cli)

    yield cli

    client_cleanup(cli)

    # No logout yet. Once it is implemented,
    # here a logout must be posted.



"""
The Client class is the main STELAR API client object.

The Client class primarily holds the API URL and user credentials needed to
access the API.
"""

from __future__ import annotations

from configparser import ConfigParser
from http.client import responses
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urljoin, urlparse, urlunparse

import requests
from requests.utils import get_auth_from_url, urldefragauth

from .admin import AdminAPI
from .catalog import CatalogAPI

# Import subAPIs modules
from .endpoints import APIEndpointsV1
from .knowgraph import KnowledgeGraphAPI
from .s3 import S3API
from .wfapi import WorkflowsAPI

if TYPE_CHECKING:
    from os import PathLike


class Client(WorkflowsAPI, CatalogAPI, KnowledgeGraphAPI, AdminAPI, S3API):
    """An SDK (client) for the STELAR API.

    Operation of the STELAR client requires three pieces of information:
        1. The base URL of the STELAR installation.
        2. A username for the user.
        3. Access and refresh tokens for the user. These tokens need to be refreshed periodically.

    The client can be initialized in four ways:
        1. By providing nothing. This is equivalent to specifying the context name "default".
        2. By providing a context name, which is looked up in the config file (see below).
        3. By providing a base URL, username, and password. The client will then authenticate
           the user and retrieve the access and refresh tokens. Note that, the password is
           not stored in the client or anywhere else.
        4. By providing a base URL, username, password, and a token JSON dictionary containing
           the access token, refresh token, and their expiration times.

    When options 3. or 4. above are used, when the tokens expire, the user must call
    either the method `reauthenticate(passwd)` providing the password, or calling method
    `reset_tokens(token_json)` with a new token JSON dictionary. Again, these tokens will
    eventually expire.

    By contrast, when the client is initialized with a context name, the client
    automatically reacquires the tokens when they expire, since the context config file
    contains the password needed to re-authenticate the user.

    By default the config file is located at $HOME/.stelar, but can be overridden by
    providing the 'config_file' keyword argument. The config file contains a
    collection of contexts, and is encoded in the INI format:

    | [default]
    | base_url=https://klms.example.com
    | username=joe
    | password=my!secret
    |
    | [admin]
    | base_url=https://klms.example.com
    | username=admin
    | password=very!secret

    The token_json dictionary is expected to have the following (indicative) structure:

    | {
    |     "access_token": "your_access_token",
    |     "refresh_token": "your_refresh_token",
    |     "expires_in": 600,
    |     "refresh_expires_in": 7200,
    | }


    Args:
        context (str): load the specified context from $HOME/.stelar (or
            the `config_file` path). If this is None, the default context is used.
            If a config file is not found, the base_url, username and password
            can be provided as keywords.

        base_url (str): The base URL to the STELAR installation. This URL contains only the
            hostname. Optionally, it may contain a user name and password, as in
            https://joe:joespassword@klms.example.com/
            The user name and password are only used if the keyword arguments 'username' and
            'password' are None.

        username (str): The user name to connect to for this client.
        password (str): The password to authenticate with.
        token_json (dict): A dictionary containing the access token, refresh token, and their expiration times.

        tls_verify (bool):  Verify the server TLS certificate. This setting takes precedence
            if given. If none, the default is to verify.

        config_file (PathLike): Path to the config file. If None, the default of "$HOME/.stelar" is
        used.
    """

    def __init__(
        self,
        context: str = None,
        *,
        base_url: str = None,
        username=None,
        password=None,
        token_json: dict = None,
        tls_verify=True,
        config_file: PathLike = None,
    ):
        # Validate base_url
        if context is None and base_url is None:
            context = "default"
        self._config_file = config_file
        self._context = context

        if not tls_verify:
            import urllib3

            urllib3.disable_warnings()

        if self._context is not None:  # init via context
            base_url, username, password = self.__from_context()
            token_json = self.authenticate(
                base_url, username=username, password=password, tls_verify=tls_verify
            )

        else:  # have base_url, get username and password
            uuser, upass = get_auth_from_url(base_url)
            if not username:
                username = uuser
            if not password:
                password = upass
            base_url = self.__normalize_base_url(base_url)
            if not token_json:
                token_json = self.authenticate(
                    base_url,
                    username=username,
                    password=password,
                    tls_verify=tls_verify,
                )

        self._username = username
        super().__init__(base_url, token_json, tls_verify)

    def reauthenticate(self, password: str = None):
        """Refresh the access token for this client.

        This method attempts to refresh the access token using the refresh token. If the
        refresh token is not available, or if it is also stale, the method will attempt to
        re-authenticate the user.

        Args:
            password (str): A password, which is used for username/password authentication.
                If not provided, the context mechanism is used.

        Raises:
            RuntimeError: If the refresh token is invalid or if re-authentication fails.
        """

        # First, try to use the refresh token.
        if self._refresh_token is not None:
            try:
                token_json = self.token_refresh(
                    self._base_url, self._refresh_token, self._tls_verify
                )
                self.reset_tokens(token_json)
                return
            except RuntimeError as e:
                # Suppress exception, proceed to refresh by re-authentication
                print(
                    "Refreshing token: an exception occurred using the refresh token:",
                    e,
                )
                pass

        if self._context:
            base_url, username, password = self.__from_context()
        else:
            username = self._username

        token_json = self.authenticate(
            self._base_url,
            username=username,
            password=password,
            tls_verify=self._tls_verify,
        )

        # Reset the client tokens
        self.reset_tokens(token_json)

    @classmethod
    def token_refresh(
        cls, base_url: str, refresh_token: str, tls_verify: bool = True
    ) -> dict:
        """
        Use the given refresh token to retrieve new access and refresh tokens.

        Args:
            base_url (str): The URL of the STELAR service
            refresh_token (str): The refresh token to use.
            tls_verify (bool): Whether to verify the server TLS certificate.

        Returns:
            A dict containing the access token, refresh token, token expiration times and the type of token
            (should be 'Bearer').

        Raises:
            RuntimeError: If authentication fails due to incorrect credentials or server issues.
        """

        req_data = {"refresh_token": refresh_token}
        req_url = urljoin(base_url, "/stelar/api/" + APIEndpointsV1.TOKEN_ISSUE)
        token_response = requests.put(
            url=req_url,
            json=req_data,
            headers={"Content-Type": "application/json"},
            verify=tls_verify,
        )
        status_code = token_response.status_code
        if status_code >= 500:  # Could be a message by the proxy, or a server error
            raise RuntimeError(
                "Could not authenticate user. The server is not available.",
                status_code,
                responses.get(status_code, "Unknown status code"),
                token_response.text,
            )
        token_json = token_response.json().get("result", None)
        success = token_response.json().get("success")

        if success and status_code in range(200, 300):
            return token_json
        else:
            raise RuntimeError(
                "Could not refresh the current token", status_code, token_json
            )

    @classmethod
    def authenticate(cls, base_url, username, password, tls_verify=True) -> dict:
        """
        Authenticates the user and retrieves access and refresh tokens.

        This method sends a POST request to the authentication endpoint using the provided
        username and password. Upon successful authentication, the method updates the token
        property and refreshes the token for all subAPI instances.

        Args:
            base_url (str): The URL of the STELAR service
            username (str): The username of the user.
            password (str): The password of the user.
            tls_verify (str): Whether to verify the server TLS certificate.

        Raises:
            RuntimeError: If authentication fails due to incorrect credentials or server issues.

        Returns:
            A dict containing the access token, refresh token, token expiration times and the type of token
            (should be 'Bearer').
        """

        auth_data = {"username": username, "password": password}
        req_url = urljoin(base_url, "/stelar/api/" + APIEndpointsV1.TOKEN_ISSUE)

        token_response = requests.post(
            url=req_url,
            json=auth_data,
            headers={"Content-Type": "application/json"},
            verify=tls_verify,
        )
        status_code = token_response.status_code
        if status_code >= 500:  # Could be a message by the proxy, or a server error
            raise RuntimeError(
                "Could not authenticate user. The server is not available.",
                status_code,
                responses.get(status_code, "Unknown status code"),
                token_response.text,
            )

        js = token_response.json()

        if status_code == 200:
            token_json = js["result"]
            return token_json
        else:
            raise RuntimeError(
                "Could not authenticate user.",
                status_code,
                js,
            )

    @staticmethod
    def __normalize_base_url(base_url):
        """
        Returns the base url for the STELAR service.

        This is computed from any partial URL, by appending
        the default path prefix '/stelar'. Also, any fragment
        or authentication info is removed.

        Args:
            base_url (str): The base URL to normalize.

        Returns:
            str: The API URL.
        """
        burl = urldefragauth(base_url)
        return urljoin(burl, "stelar")

    def __from_context(self):
        config_file = (
            self._config_file if self._config_file else Path.home() / ".stelar"
        )
        context = self._context
        c = ConfigParser()
        c.read(config_file)
        if not c.has_section(context):
            raise ValueError(f"Client context '{context}' does not exist")
        ctx = c[context]
        base_url = self.__normalize_base_url(ctx["base_url"])
        usr = ctx["username"]
        pwd = ctx["password"]
        self._ckan_apitoken = ctx.get("ckan_apitoken", None)
        return base_url, usr, pwd

    @property
    def DC(self):
        try:
            return self._ckan_client
        except AttributeError:
            from .backdoor import CKAN

            self._ckan_client = CKAN(client=self)
            return self._ckan_client

    __repr_classname = "stelar.client.Client"

    def __repr__(self):
        purl = urlparse(self._base_url)
        if self._username:
            netloc = f"{self._username}@{purl.netloc}"
        else:
            netloc = purl.netloc
        enhanced_url = urlunparse((purl.scheme, netloc, purl.path, "", "", ""))
        return f"{self.__repr_classname}({enhanced_url})"



# ----------------------------------------------------------
# Exceptions raised by the proxy system.
# ----------------------------------------------------------
__all__ = [
    "EntityError",
    "ConversionError",
    "ProxyError",
    "InvalidationError",
    "ConflictError",
    "ErrorState",
    "ProxyOperationError",
    "EntityNotFound",
]


class EntityError(ValueError):
    """Raised when the system encounters a malformed entity"""

    pass


class ConversionError(EntityError):
    """A problem with conversion to and from entity"""

    def __init__(self, property, conv_type):
        super().__init__(
            f"Conversion failed for {property} {conv_type}", property.name, conv_type
        )


class ProxyError(RuntimeError):
    """The proxying state is not consistent or an illegal proxying command was issued"""

    pass


class ConflictError(ProxyError):
    """Update for an entity with unsynchronized changes"""

    def __init__(self, conflicted_proxy, new_entity, *args):
        super().__init__(*args)
        self.conflicted_proxy = conflicted_proxy
        self.new_entity = new_entity


class InvalidationError(ProxyError):
    """Invalidation attempt on dirty proxy"""

    pass


class ErrorState(ProxyError):
    """Operation on proxy in ERROR state"""

    pass


class ProxyOperationError(ProxyError):
    """An error occurred during an API operation"""

    def __init__(self, /, proxy_type, eid, operation, *args, **kwargs):
        if isinstance(proxy_type, type):
            self.proxy_type = proxy_type.__name__
        else:
            self.proxy_type = str(proxy_type)
        self.eid = str(eid)
        self.operation = operation
        super().__init__(*args, **kwargs)

    def __repr__(self):
        typename = self.proxy_type.__name__
        return f"{self.__class__.__name__}({self.operation} {typename} {self.eid} {self.args})"


class EntityNotFound(ProxyOperationError):
    """Indicate that an entity is not found"""

    def __init__(self, /, proxy_type, eid, operation):
        super().__init__(proxy_type, eid, operation)



from uuid import uuid4

from proxy_utils import ProxyTestObj, TPCatalog

from stelar.client.proxy import Id, Property, Proxy, RefList, Registry


def test_simple_subset_decl():
    class Foo(ProxyTestObj):
        a = Property()

    class Bar(ProxyTestObj):
        b = Property()
        acoll = RefList(Foo)

    assert "acoll" in Bar.proxy_schema.properties
    acoll_descriptor = Bar.proxy_schema.properties["acoll"]
    assert acoll_descriptor.name == "acoll"
    assert acoll_descriptor.proxy_type is Foo


def test_collection_init():
    uuids = [uuid4() for i in range(4)]

    class Foo(ProxyTestObj):
        a = Property()

    Foo.data = {uuids[i]: {"id": uuids[i], "a": f"this is a {i}"} for i in range(3)}

    class Bar(ProxyTestObj):
        b = Property()
        acoll = RefList(Foo)

    Bar.data = {
        uuids[3]: {
            "id": uuids[3],
            "b": "this is b",
            "acoll": [Foo.data[u] for u in uuids[:3]],
        }
    }

    c = TPCatalog()
    foo_cache = c.registry_for(Foo)
    bar_cache = c.registry_for(Bar)



"""
Proxy field validation
----------------------

Proxied entity fields (attributes) have two representations: one
that appears in the entity (JSON object stored in a dict) and one that is used
by the proxy objects.

client proxy value  <-->  json entity value

For example, dates are represented as datetime objects in the proxy and as strings
in the JSON entity. Conversion between the two is done via the two conversion methods
of the classes herein.
"""

from __future__ import annotations

import re
from collections.abc import MutableSequence
from datetime import datetime
from typing import Any, Mapping, Optional
from uuid import UUID

__all__ = [
    "FieldValidator",
    "AnyField",
    "EnumeratedField",
    "StateField",
    "BoolField",
    "IntField",
    "StrField",
    "DateField",
    "DateFmtField",
    "DictField",
    "ListField",
    "BasicField",
    "ExecStateField",
    "UUIDField",
    "NameField",
    "TagNameField",
    "VocabNameField",
]


class FieldValidator:
    """Provide simple validation and conversion for entity fields.

        A validation is a sequence of checks. Each check can:
        - Raise an exception of ValueError
        - Apply a value conversion and continue
        - Apply a value conversion and terminate

    Any function that takes as input a value, and returns a pair (newvalue, done)
    where `done` is a boolean, can be used as a check.

    At the end of all conversions, if no conversion signaled `done`, the `strict` attribute
    is checked. If True, an error is raised, else (the default) conversion succeeds.
    """

    def __init__(
        self,
        *,
        strict: bool = False,
        nullable: bool = True,
        default: Any = ...,
        minimum_value: Any = None,
        maximum_value: Any = None,
        maximum_len: Optional[int] = None,
        minimum_len: Optional[int] = None,
    ):
        self.prioritized_checks = []
        self.checks = []
        self.strict = strict
        self.nullable = nullable
        if default is not ...:
            self.default = default
        self.minimum_value = minimum_value
        self.maximum_value = maximum_value
        self.maximum_len = maximum_len
        self.minimum_len = minimum_len

        if nullable is not None:
            self.add_check(self.check_null, -1)

        if minimum_value is not None:
            self.add_check(self.check_minimum, 10)
        if maximum_value is not None:
            self.add_check(self.check_maximum, 12)
        if maximum_len is not None or minimum_len is not None:
            self.add_check(self.check_length, 20)

    def add_check(self, check_func, pri: int):
        self.prioritized_checks.append((check_func, pri))
        self.prioritized_checks.sort(key=lambda p: p[1])
        self.checks = [p[0] for p in self.prioritized_checks]

    def check_null(self, value, **kwargs):
        if value is None:
            if self.nullable:
                return None, True
            else:
                raise ValueError("None is not allowed")
        else:
            return value, False

    def check_length(self, value, **kwargs):
        value_len = len(value)
        if self.minimum_len is not None and value_len < self.minimum_len:
            raise ValueError(
                f"The length ({value_len}) is less than the minimum ({self.minimum_len})"
            )
        if self.maximum_len is not None and value_len > self.maximum_len:
            raise ValueError(
                f"The length ({value_len}) is greater that the maximum ({self.maximum_len})"
            )
        return value, False

    def check_minimum(self, value, **kwargs):
        if value < self.minimum_value:
            raise ValueError(f"Value ({value}) too low (minimum={self.minimum_value})")
        return value, False

    def check_maximum(self, value, **kwargs):
        if value > self.maximum_value:
            raise ValueError(f"Value ({value}) too high (maximum={self.maximum_value})")
        return value, False

    def validate(self, value, **kwargs):
        done = False
        try:
            for check in self.checks:
                value, done = check(value, **kwargs)
                if done:
                    return value
        except ValueError:
            raise
        except Exception as e:
            raise ValueError("Bad value found during validation") from e

        if self.strict:
            raise ValueError("Validation failed to match input value")
        else:
            return value

    def convert_to_proxy(self, value, **kwargs):
        raise NotImplementedError()

    def convert_to_entity(self, value, **kwargs):
        raise NotImplementedError()

    def default_value(self):
        raise NotImplementedError

    def repr_constraints(self):
        nn = ["nullable" if self.nullable else "not null"]

        if self.minimum_len is not None and self.maximum_len is not None:
            nn.append(f"{self.minimum_len} <= length <= {self.maximum_len}")
        elif self.minimum_len is not None:
            nn.append(f"{self.minimum_len} <= length")
        elif self.maximum_len is not None:
            nn.append(f"length <= {self.maximum_len}")

        if self.minimum_value is not None and self.maximum_value is not None:
            nn.append(f"{self.minimum_value} <= value <={self.maximum_value}")
        elif self.minimum_value is not None:
            nn.append(f"{self.minimum_value} <= value")
        elif self.maximum_value is not None:
            nn.append(f"value <= {self.maximum_value}")

        return nn


class AnyField(FieldValidator):
    """A very promiscuous basic validator."""

    def __init__(self, repr_type="Any", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._repr_type = repr_type

    def convert_to_proxy(self, value, **kwargs):
        return value

    def convert_to_entity(self, value, **kwargs):
        return value

    def default_value(self):
        if hasattr(self, "default"):
            return self.default
        elif self.nullable:
            return None
        else:
            raise NotImplementedError()

    def repr_type(self):
        return self._repr_type


class EnumeratedField(AnyField):
    """Fields with a fixed set of legal values.

    Subclasses can redefine the VALUES class attribute.
    """

    VALUES = []

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.add_check(self.oneof, 5)

    def oneof(self, value):
        if value not in self.VALUES:
            raise ValueError(f"Not one of {self.VALUES}")
        return value, False

    def repr_type(self):
        return f"OneOf{self.VALUES}"


class StateField(EnumeratedField):
    VALUES = ["active", "deleted"]

    def __init__(self):
        super().__init__(nullable=False)


class ExecStateField(EnumeratedField):
    VALUES = ["running", "succeeded", "failed"]

    def __init__(self):
        super().__init__(nullable=False)


class BasicField(AnyField):
    """
    Given ftype T, accept value if it is an instance of T or if T(value) succeeds.
    Conversion to T is performed.

    Subclasses include basic types: str, int, bool
    """

    def __init__(self, ftype, **kwargs):
        super().__init__(**kwargs)
        self.ftype = ftype
        self.add_check(self.to_ftype, 5)

    def to_ftype(self, value, **kwargs):
        """Validator stage for BasicField"""
        if not isinstance(value, self.ftype):
            value = self.ftype(value)
        return value, False

    def repr_type(self):
        return self.ftype.__name__


class StrField(BasicField):
    """A string field validator"""

    def __init__(self, **kwargs):
        super().__init__(ftype=str, **kwargs)


class NameField(StrField):
    """Name fields are non-nullable string fields whose value
    must follow a pattern.
    """

    def __init__(self, **kwargs):
        super().__init__(nullable=False, minimum_len=2, maximum_len=100, **kwargs)
        self.add_check(self.check_name, 7)

    NAME_PATTERN = re.compile(r"[a-z0-9_-]+")

    def check_name(self, value: str, **kwargs):
        if self.NAME_PATTERN.fullmatch(value) is None:
            raise ValueError(
                f"Name must be a string matching '{self.NAME_PATTERN.pattern}'"
            )
        return value, False


class VocabNameField(NameField):
    NAME_PATTERN = re.compile(r".+")


class TagNameField(NameField):
    NAME_PATTERN = re.compile(r"[A-Za-z0-9 ._-]+")


class IntField(BasicField):
    """An int field validator"""

    def __init__(self, **kwargs):
        super().__init__(ftype=int, **kwargs)


class BoolField(BasicField):
    """A bool field validator"""

    def __init__(self, **kwargs):
        super().__init__(ftype=bool, **kwargs)


class DateField(FieldValidator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_check(self.to_date, 5)

    def to_date(self, value: Any, **kwargs) -> tuple[datetime, bool]:
        """Validation stage for dates."""
        if isinstance(value, str):
            return datetime.fromisoformat(value), False
        elif isinstance(value, datetime):
            return value, False
        else:
            raise ValueError("Invalid type, expected datetime or string")

    def convert_to_entity(self, value: datetime, **kwargs) -> str:
        return value.isoformat()

    def convert_to_proxy(self, value: str, **kwargs) -> datetime:
        return datetime.fromisoformat(value)

    def repr_type(self):
        return "datetime"


class DateFmtField(FieldValidator):
    def __init__(self, datefmt, **kwargs):
        super().__init__(**kwargs)
        self.datefmt = datefmt
        self.add_check(self.to_date, 5)

    def to_date(self, value: Any, **kwargs) -> tuple[datetime, bool]:
        """Validation stage for dates."""
        if isinstance(value, str):
            return datetime.strptime(value, self.datefmt), False
        elif isinstance(value, datetime):
            return value, False
        else:
            raise ValueError("Invalid type, expected datetime or string")

    def convert_to_entity(self, value: datetime, **kwargs) -> str:
        return value.strftime(self.datefmt)

    def convert_to_proxy(self, value: str, **kwargs) -> datetime:
        return datetime.strptime(value, self.datefmt)

    def repr_type(self):
        return "datetime"


class DictField(AnyField):
    def __init__(self, key_type, value_type, **kwargs):
        super().__init__(**kwargs)
        self.key_type = key_type
        self.value_type = value_type
        self._repr_type = f"Dict[{key_type},{value_type}]"
        self.add_check(self.check_dict, 5)

    def check_dict(self, value, **kwargs):
        if not isinstance(value, Mapping):
            raise TypeError("Expected a dictionary")
        for k, v in value.items():
            if not isinstance(k, self.key_type):
                raise TypeError(f"Invalid key type, expected {self.key_type}")
            if not isinstance(v, self.value_type):
                raise TypeError(f"Invalid value type, expected {self.value_type}")
        return dict(value), False


class ListField(AnyField):
    def __init__(self, element_type, **kwargs):
        super().__init__(**kwargs)
        self.element_type = element_type
        self._repr_type = f"List[{element_type}]"
        self.add_check(self.check_list, 5)

    def check_list(self, value, **kwargs):
        if not isinstance(value, MutableSequence):
            raise TypeError("Expected a list")
        for v in value:
            if not isinstance(v, self.element_type):
                raise TypeError(f"Invalid element type, expected {self.element_type}")
        return list(value), False

    def repr_type(self):
        return self._repr_type


class UUIDField(BasicField):
    def __init__(self, **kwargs):
        super().__init__(ftype=UUID, **kwargs)

    def convert_to_proxy(self, value: str, **kwargs) -> UUID:
        try:
            return UUID(value)
        except ValueError:
            raise ValueError("Invalid UUID", value, kwargs)

    def convert_to_entity(self, value: UUID, **kwargs) -> str:
        return str(value)

    def repr_type(self):
        return "UUID"



from stelar.client import Client, Process, Workflow


# @pytest.fixture(autouse=True, scope="module")
def cleanup_processes(testcontext):
    c = Client(testcontext)

    # Run after the tests in the module
    yield None

    for p in c.processes[:]:
        try:
            if "test_delete" in p.tags:
                p.delete()
                assert p not in c.processes
                c.DC.dataset_purge(id=str(p.id))
        except Exception:
            pass


def test_process_create(testcli):
    p = Process.new(testcli)
    assert p.organization is testcli.organizations["stelar-klms"]
    assert p.exec_state == "running"
    assert p.state == "active"
    assert p.creator == "admin"

    assert p.tasks == []

    p.terminate("succeeded")
    assert p.exec_state == "succeeded"
    p.tags += ("test_delete",)
    p.delete()


def test_process_create_with_workflow(testcli):
    wf = Workflow.new(testcli, name="test_workflow")

    p = Process.new(testcli, workflow=wf)
    assert p.workflow is wf
    assert p.exec_state == "running"
    assert p.state == "active"
    assert p.creator == "admin"

    assert p.tasks == []

    assert p.workflow is wf

    # Check updates to workflow
    p.workflow = None
    assert p.workflow is None
    p.workflow = wf
    assert p.workflow is wf

    p.terminate("succeeded")
    assert p.exec_state == "succeeded"
    p.tags += ("test_delete",)
    p.delete()
    wf.delete(purge=True)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from .api_call import api_call
from .generic import GenericCursor, GenericProxy
from .package import PackageCursor, PackageProxy
from .proxy import (
    DateFmtField,
    EnumeratedField,
    Id,
    Property,
    StrField,
    derived_property,
)
from .proxy.property import DictProperty
from .task_spec import TaskSpec


class ToolCategoryField(EnumeratedField):
    VALUES = [
        "discovery",
        "interlinking",
        "annotation",
        "other",
    ]


#    {'is_manifest_list': False,                                            │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$ git commit -a
#    'last_modified': 'Sun, 01 Jun 2025 19:45:45 -0000',                             │[main e7923b9] Fixed the proxy resync code related to task creation and termination.
#    'manifest_digest': 'sha256:6e4a55d6531c1aa8844164b552ba7b5f5ef9b30c5d9d94a145777│ 4 files changed, 49 insertions(+), 4 deletions(-)
# c7d26b7c572',                                                                       │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$
#    'name': '0.2.0',                                                                │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$
#    'reversion': False,                                                             │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$ git pull
#    'size': 56837148,                                                               │Already up to date.
#    'start_ts': 1748807145
# }


@dataclass
class Image:
    """A container image used by a tool.

    Images are just information objects, they are not proxies.
    """

    is_manifest_list: bool
    last_modified: datetime
    manifest_digest: str
    tool: Tool  # The tool this image belongs to
    name: str
    reversion: bool
    size: int
    start_ts: datetime

    def __post_init__(self):
        if isinstance(self.last_modified, str):
            self.last_modified = datetime.strptime(
                self.last_modified, "%a, %d %b %Y %H:%M:%S -0000"
            )
        if isinstance(self.start_ts, int):
            self.start_ts = datetime.fromtimestamp(self.start_ts)

    @property
    def id(self):
        """Return a unique identifier for this image."""
        return f"stelar/{self.tool.name}:{self.name}"

    def __repr__(self):
        return f"Image({self.tool.name}/{self.name})"


class Tool(PackageProxy):
    # weird ones
    # license_id = Property(validator=StrField(nullable=True), updatable=True)
    git_repository = Property(validator=StrField(nullable=True), updatable=True)
    programming_language = Property(validator=StrField(nullable=True), updatable=True)
    version = Property(
        validator=StrField(nullable=True, maximum_len=100), updatable=True
    )

    inputs = DictProperty(str, str, updatable=True)
    outputs = DictProperty(str, str, updatable=True)
    parameters = DictProperty(str, str, updatable=True)

    category = Property(validator=ToolCategoryField(nullable=True), updatable=True)

    repository = Property(validator=StrField(nullable=False), updatable=False)

    @derived_property
    def images(self, entity):
        """Return a list of images associated with this tool."""
        images = entity.get("images", [])
        return [Image(tool=self, **image) for image in images]

    def task_spec(self, image=None):
        """Return a new TaskSpec initialized for this tool"""
        return TaskSpec(tool=self, image=image)


class ToolCursor(PackageCursor[Tool]):
    def __init__(self, client):
        super().__init__(client, Tool)


class ImageRegistryToken(GenericProxy):
    """A token representing an image used by a tool."""

    DATEFORMAT = "%a, %d %b %Y %H:%M:%S %z"

    id = Id(entity_name="uuid")
    created = Property(validator=DateFmtField(DATEFORMAT), updatable=False)
    expiration = Property(
        validator=DateFmtField(DATEFORMAT, nullable=True), updatable=False
    )
    last_accessed = Property(
        validator=DateFmtField(DATEFORMAT, nullable=True), updatable=False
    )

    title = Property(validator=StrField(nullable=False), updatable=False)

    token = Property(
        validator=StrField(nullable=False), updatable=False, entity_name="token_code"
    )


class ImageRegistryTokenCursor(GenericCursor[ImageRegistryToken]):
    """A cursor for iterating over image registry tokens."""

    def __init__(self, client):
        super().__init__(client, ImageRegistryToken)

    def create(
        self, title: str, expiration: datetime | None = None
    ) -> ImageRegistryToken:
        """Create a new image registry token."""
        ac = api_call(self)
        result = ac.image_registry_token_create(title=title, expiration=expiration)
        return self.fetch_proxy_for_entity(result)



from __future__ import annotations

import pytest

from stelar.client import TaskSpec


@pytest.mark.skip()
def test_exec_word_count_manual(test_client):
    """
    Execute the word count task manually.
    """

    c = test_client
    d1 = c.datasets["shakespeare_novels"]

    # Create a process
    p1 = c.processes.get("word-count-tests")

    api_json = {
        "process_id": str(p1.id),
        # name: Optional, goes into tags
        "name": "Word Count on RJ",
        # Package name or id of the tool
        # If image is not provided, take the 'latest' (most recent) image of the tool
        # from the embedded repository
        # If image is provided use the image
        ### "tool": ""
        "image": "petroud/stelar-word-count:latest",
        #
        "inputs": {"text_to_count": ["d1::owned"]},
        "datasets": {
            "d1": str(d1.id),
            "d2": {
                "name": "the-results-of-word-count",
                "owner_org": "stelar-klms",
                "notes": "The result of map reduce word count",
                "tags": ["wordcount", "map-reduce"],
            },
        },
        # Use anything, not used
        "parameters": {"key": "value", "key1": 1},
        "outputs": {
            "word_count_file": {
                "url": "s3://test-bucket/word-count.txt",
                "dataset": "d2",
                "resource": {"name": "RJ Word Count 2", "relation": "owned"},
            }
        },
        "secrets": {"hello": "world"},
    }

    # Define the input and output files
    resp = c.POST("v2/task", **api_json)

    assert resp.status_code in range(
        200, 300
    ), f"Error: {resp.status_code} - {resp.text}"
    assert (
        resp.json()["status"] == "success"
    ), f"Error: {resp.status_code} - {resp.text}"


def test_task_spec(testcli):
    d1 = testcli.datasets["dataset1"]

    t = TaskSpec(name="Word Count on RJ")
    t.d("d1", d1)
    t.d(
        "d2",
        name="the-results-of-word-count",
        owner_org="stelar-klms",
        notes="The result of map reduce word count",
        tags=["wordcount", "map-reduce"],
    )
    t.i(text_to_count="d1::owned", profile=["d1::profile"])
    t.o(
        word_count_file={
            "url": "s3://test-bucket/word-count.txt",
            "dataset": "d2",
            "resource": {"name": "RJ Word Count 2", "relation": "owned"},
        }
    )
    t.p(key="value", key1=1)

    spec = t.spec()
    target = {
        "name": "Word Count on RJ",
        "datasets": {
            "d1": str(d1.id),
            "d2": {
                "name": "the-results-of-word-count",
                "owner_org": "stelar-klms",
                "notes": "The result of map reduce word count",
                "tags": ["wordcount", "map-reduce"],
            },
        },
        "inputs": {"text_to_count": ["d1::owned"], "profile": ["d1::profile"]},
        "outputs": {
            "word_count_file": {
                "url": "s3://test-bucket/word-count.txt",
                "dataset": "d2",
                "resource": {"name": "RJ Word Count 2", "relation": "owned"},
            }
        },
        "parameters": {
            "key": "value",
            "key1": 1,
        },
    }

    assert set(spec.keys()) == set(target.keys())
    for key in spec:
        assert spec[key] == target[key]


def test_task_spec_empty():
    t = TaskSpec()
    spec = t.spec()
    target = {
        "name": "unnamed",
        "datasets": {},
        "inputs": {},
        "outputs": {},
        "parameters": {},
    }

    assert set(spec.keys()) == set(target.keys())
    for key in spec:
        assert spec[key] == target[key]


def test_task_spec_init(testcli):
    """
    Test the initialization of TaskSpec.
    """

    assert TaskSpec().tool is None
    assert TaskSpec().name is None
    assert TaskSpec().image is None

    assert TaskSpec("test-tool").tool == "test-tool"

    tool = testcli.tools["simple_tool"]
    assert TaskSpec(tool).tool == str(tool.id)
    assert TaskSpec(tool.id).tool == str(tool.id)
    assert TaskSpec(tool.name).tool == tool.name

    assert TaskSpec(tool).image is None
    assert TaskSpec(tool, image="latest").image == "latest"
    assert (
        TaskSpec(image="petroud/stelar-word-count:latest").image
        == "petroud/stelar-word-count:latest"
    )
    assert TaskSpec(image="petroud/stelar-word-count:latest").tool is None


def test_task_spec_dataset(testcli):
    """
    Test the dataset method of TaskSpec.
    """

    d1 = testcli.datasets["dataset1"]
    t = TaskSpec().d("d1", d1)
    assert t.datasets == {"d1": str(d1.id)}

    t = TaskSpec().d("d2", name="test-dataset", owner_org="stelar-klms")
    assert t.datasets == {"d2": {"name": "test-dataset", "owner_org": "stelar-klms"}}

    with pytest.raises(ValueError):
        t = TaskSpec().d("d3", d1, name="test-dataset", owner_org="stelar-klms")

    deforg = testcli.datasets.default_organization
    assert TaskSpec().d("d", name="foo", owner_org=deforg).datasets["d"][
        "owner_org"
    ] == str(deforg.id)

    assert TaskSpec().d("d", name="foo", owner_org=deforg.id).datasets["d"][
        "owner_org"
    ] == str(deforg.id)


def test_task_spec_input(testcli):
    """
    Test the input method of TaskSpec.
    """

    d = testcli.datasets["shakespeare_novels"]
    r = d.resources[0]
    rid = str(r.id)

    assert TaskSpec().i(inp=r).spec()["inputs"]["inp"] == [rid]
    assert TaskSpec().i(inp=rid).spec()["inputs"]["inp"] == [rid]

    assert TaskSpec().i(inp=[r]).spec()["inputs"]["inp"] == [rid]
    assert TaskSpec().i(inp=[rid]).spec()["inputs"]["inp"] == [rid]

    assert TaskSpec().i(inp=[r, r]).spec()["inputs"]["inp"] == [rid, rid]
    assert TaskSpec().i(inp=[rid, rid]).spec()["inputs"]["inp"] == [rid, rid]
    assert TaskSpec().i(inp=[r, rid]).spec()["inputs"]["inp"] == [rid, rid]
    assert TaskSpec().i(inp=[rid, r]).spec()["inputs"]["inp"] == [rid, rid]


def test_task_spec_output(testcli):
    """
    Test the output method of TaskSpec.
    """

    d = testcli.datasets["shakespeare_novels"]
    r = d.resources[0]
    rid = str(r.id)

    outspec = {"url": r.url}
    assert TaskSpec().o(out=outspec).spec()["outputs"]["out"] == {
        "url": r.url,
    }

    outspec = {"url": r.url, "resource": r}
    assert TaskSpec().o(out=outspec).spec()["outputs"]["out"] == {
        "url": r.url,
        "resource": str(r.id),
    }

    outspec = {
        "url": r.url,
        "dataset": str(d.id),
        "resource": {"name": r.name, "relation": "owned"},
    }
    assert TaskSpec().o(out=outspec).spec()["outputs"]["out"] == {
        "url": r.url,
        "dataset": str(d.id),
        "resource": {"name": r.name, "relation": "owned"},
    }


#
# TODO: Add more tests for task execution
#
def test_create_task(testcli):
    """
    Test the creation of a task.
    """

    proc = testcli.processes["simple_proc"]
    task = proc.run(TaskSpec())

    assert task.exec_state == "created"
    assert task.process is proc
    assert task.inputs == {}

    task.exit_job()
    assert task.exec_state == "succeeded"

    task.delete()


def test_create_task_with_fail(testcli):
    """
    Test the creation of a task that fails.
    """

    proc = testcli.processes["simple_proc"]
    task = proc.run(TaskSpec())

    assert task.exec_state == "created"
    assert task.process is proc
    assert task.inputs == {}
    assert task.parameters is None

    assert not hasattr(task, "outputs")
    assert not hasattr(task, "messages")

    # Simulate a failure
    task.fail("Simulated failure")

    assert task.exec_state == "failed"
    assert task.messages == "Simulated failure"

    task.delete()


def test_create_task_params(testcli):
    """
    Test the creation of a task with parameters.
    """

    proc = testcli.processes["simple_proc"]
    task = proc.run(
        TaskSpec().p(
            str_param="test",
            int_param=42,
            composite={"key": "value"},
            listparam=["item1", "item2"],
        )
    )
    assert task.exec_state == "created"
    assert task.process is proc
    assert task.inputs == {}
    assert task.parameters == {
        "str_param": "test",
        "int_param": 42,
        "composite": {"key": "value"},
        "listparam": ["item1", "item2"],
    }

    task.exit_job()
    task.delete()


def test_create_task_with_inputs(testcli):
    # Testing for different types of input declarations

    c = testcli
    proc = c.processes["simple_proc"]

    shakespeare = c.datasets["shakespeare_novels"]
    iris = c.datasets["iris"]
    wine = c.datasets["wine"]

    ts = TaskSpec()

    ts.d("wine_quality", wine)

    # Add from resource
    ts.i(iris=iris.resources[0])

    # Add from resource id
    ts.i(wine=str(wine.resources[0].id))

    # Add from dataset and relation
    ts.i(shakespeare=f"{shakespeare.id}::owned")

    # Add from dataset and relation with alias
    ts.i(wine_alias="wine_quality::owned")

    # Add raw url
    ts.i(raw_url="s3://klms-bucket/iris.csv")

    # Add url from process context
    ts.i(ctx_file="ctx::testing")

    t = proc.run(ts)
    assert t.exec_state == "created"
    tin = t.job_input

    assert tin["input"]["iris"] == [str(iris.resources[0].url)]
    assert tin["input"]["wine"] == [str(wine.resources[0].url)]
    assert tin["input"]["shakespeare"] == [str(shakespeare.resources[0].url)]
    assert tin["input"]["wine_alias"] == [str(wine.resources[0].url)]
    assert tin["input"]["raw_url"] == ["s3://klms-bucket/iris.csv"]
    assert tin["input"]["ctx_file"] == ["s3://klms-bucket/wordcount.csv"]

    t.exit_job()
    t.delete()


def test_create_task_with_outputs(testcli):
    """
    Test the creation of a task with outputs.
    """

    c = testcli
    proc = c.processes["simple_proc"]
    stelar_klms = c.organizations["stelar-klms"]

    if "test_dataset" in c.datasets:
        c.datasets["test_dataset"].delete(purge=True)
    if "output_dataset" in c.datasets:
        c.datasets["output_dataset"].delete(purge=True)

    test_dataset = c.datasets.create(
        name="test_dataset", organization=stelar_klms, title="Test Dataset"
    )
    test_dataset.add_resource(
        name="Test resource0", url="s3://klms-bucket/badfile0.csv"
    )
    test_dataset.add_resource(
        name="Test resource1", url="s3://klms-bucket/badfile1.csv"
    )
    test_dataset.add_resource(
        name="Test resource2", url="s3://klms-bucket/badfile2.csv"
    )

    ts = TaskSpec()
    ts.d("output_dataset", name="output_dataset", owner_org="stelar-klms")
    ts.d("test_dataset", test_dataset)

    ts.o(
        outfile1={
            "url": "s3://klms-bucket/output1.csv",
            "dataset": "output_dataset",
            "resource": {
                "name": "Output Resource",
                "relation": "owned",
            },
        },
        outfile2={
            "url": "s3://klms-bucket/output2.csv",
            "dataset": "ctx",
            "resource": {
                "name": "Output Resource 2",
                "relation": "temporary",
            },
        },
        outfile3={
            "url": "s3://klms-bucket/output3.csv",
            "resource": test_dataset.resources[0],
        },
        outfile4={
            "url": "s3://klms-bucket/output4.csv",
            "resource": test_dataset.resources[1].id,
        },
        outfile5={
            "url": "s3://klms-bucket/output5.csv",
            "resource": str(test_dataset.resources[2].id),
        },
        outfile6={
            "url": "s3://klms-bucket/output6.csv",
        },
    )

    proc_numres = len(proc.resources)
    t = proc.run(ts)
    assert t.exec_state == "created"
    targs = t.job_input

    assert targs["output"]["outfile1"] == "s3://klms-bucket/output1.csv"
    assert targs["output"]["outfile2"] == "s3://klms-bucket/output2.csv"
    assert targs["output"]["outfile3"] == "s3://klms-bucket/output3.csv"
    assert targs["output"]["outfile4"] == "s3://klms-bucket/output4.csv"
    assert targs["output"]["outfile5"] == "s3://klms-bucket/output5.csv"
    assert targs["output"]["outfile6"] == "s3://klms-bucket/output6.csv"

    # Create the data files
    fs = c.s3fs()
    for oname, ourl in targs["output"].items():
        fs.pipe_file(ourl, "This is the test file for " + oname)

    t.exit_job(message="Test task completed successfully", output=targs["output"])

    assert t.exec_state == "succeeded"

    # Check the resource for output1
    assert "output_dataset" in c.datasets
    output_dataset = c.datasets["output_dataset"]
    assert len(output_dataset.resources) == 1
    assert output_dataset.resources[0].url == "s3://klms-bucket/output1.csv"
    assert output_dataset.resources[0].name == "Output Resource"
    assert output_dataset.resources[0].relation == "owned"

    # Check the resource for output2
    assert len(proc.resources) == (proc_numres + 1)
    assert proc.resources[1].url == "s3://klms-bucket/output2.csv"
    assert proc.resources[1].name == "Output Resource 2"
    assert proc.resources[1].relation == "temporary"

    # Check the resource for output3
    assert len(test_dataset.resources) == 3
    assert test_dataset.resources[0].url == "s3://klms-bucket/output3.csv"
    assert test_dataset.resources[0].name == "Test resource0"

    # Check the resource for output4
    assert test_dataset.resources[1].url == targs["output"]["outfile4"]
    assert test_dataset.resources[1].name == "Test resource1"

    # Check the resource for output5
    assert test_dataset.resources[2].url == targs["output"]["outfile5"]
    assert test_dataset.resources[2].name == "Test resource2"

    # Delete catalog entries
    t.delete()
    proc.resources[1].delete()
    output_dataset.delete(purge=True)
    test_dataset.delete(purge=True)

    # Delete the output files
    for ourl in targs["output"].values():
        fs.rm(ourl)


def test_create_task_with_metrics(testcli):
    """
    Test the creation of a task with metrics.
    """

    c = testcli
    proc = c.processes["simple_proc"]

    ts = TaskSpec()
    t = proc.run(ts)
    assert t.exec_state == "created"

    mymetrics = dict(
        metric_obj={
            "value": 42,
            "description": "The answer to life, the universe, and everything",
        },
        metric_vec=[3.14, 2.71, 1.41],
        metric_str="a string",
        metric_num=12345,
        metric_null=None,
        metric_bool=True,
    )

    t.exit_job(
        message="Test task with metrics completed successfully", metrics=mymetrics
    )

    for metric in mymetrics:
        t.metrics[metric] == mymetrics[metric]
    t.delete()


def test_create_task_with_secrets(testcli):
    """
    Test the creation of a task with secrets.
    """

    c = testcli
    proc = c.processes["simple_proc"]

    ts = TaskSpec()
    secrets = {"secret_key": "secret_value"}
    t = proc.run(ts, secrets=secrets)
    assert t.exec_state == "created"

    targs = t.job_input
    assert targs["secrets"] == secrets

    t.fail("Test task with secrets failed")
    t.delete()



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from uuid import UUID

from .fieldvalidation import AnyField, UUIDField
from .property import Property
from .proxylist import ProxySublist
from .registry import Registry

if TYPE_CHECKING:
    from ..client import Client

Entity = dict[str, Any]


class RefField(AnyField):
    """This field validator is specialized for reference properties"""

    def __init__(self, ref_property: Reference, ref_typename: str, **kwargs):
        super().__init__(**kwargs)
        self.ref_property = ref_property
        self.ref_typename = ref_typename
        self.add_check(self.to_uuid, 5)

    @property
    def proxy_type(self):
        return self.ref_property.proxy_type

    def to_uuid(self, value, **kwargs):
        if not isinstance(value, self.proxy_type):
            raise ValueError(f"Expected {self.proxy_type.__name__}")
        assert isinstance(value.proxy_id, UUID)
        return value.proxy_id, True

    def convert_to_proxy(self, value: str, **kwargs) -> UUID:
        return UUID(value)

    def convert_to_entity(self, value: UUID, **kwargs) -> str:
        return str(value)

    def repr_type(self):
        return self.ref_typename


class Reference(Property):
    """A proxy property which is a reference to an entity."""

    VALIDATOR_CLASS = RefField

    def __init__(
        self,
        proxy_type,
        nullable=False,
        trigger_sync=False,
        **kwargs,
    ):
        val = self.validator_for(proxy_type, nullable=nullable)
        super().__init__(validator=val, **kwargs)
        self.__proxy_type = proxy_type
        self.nullable = nullable
        self.trigger_sync = trigger_sync

    def validator_for(self, proxy_type, **kwargs):
        ptn = self.property_type_name(proxy_type)
        return self.VALIDATOR_CLASS(self, ptn, **kwargs)

    @classmethod
    def property_type_name(cls, proxy_type) -> str:
        if isinstance(proxy_type, str):
            return proxy_type
        else:
            return proxy_type.__name__

    @property
    def proxy_type(self):
        """The class of the proxy object pointed to by this"""
        if isinstance(self.__proxy_type, str):
            from .schema import Schema

            self.__proxy_type = Schema.for_entity(self.__proxy_type).cls
        return self.__proxy_type

    def registry_for(self, obj) -> Registry:
        """Return the registry for the referrent, given owner"""
        return obj.proxy_registry.catalog.registry_for(self.proxy_type)

    def get(self, obj):
        """Low-level getter"""
        if obj.proxy_attr is None:
            obj.proxy_sync()
        if obj.proxy_attr[self.name] is None:
            return None
        else:
            return self.registry_for(obj).fetch_proxy(obj.proxy_attr[self.name])

    def __get__(self, obj, objtype=None):
        val = self.get(obj)

        # The attribute is deleted
        if val is ...:
            raise AttributeError(f"Property '{self.name}' is not present")
        return val

    # set and delete are inherited from Property


class RefListField(AnyField):
    """This field validator is specialized for reference properties"""

    def __init__(
        self,
        ref_property: RefList,
        ref_typename: str,
        *,
        element_validator=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ref_property = ref_property
        self.ref_typename = ref_typename
        self.element_validator = element_validator
        self.add_check(self.to_list, 5)

    @property
    def proxy_type(self):
        return self.ref_property.proxy_type

    def to_list(self, value, **kwargs):
        if not isinstance(value, list):
            raise ValueError(f"Expected a list")
        if self.element_validator is not None:
            value = [self.element_validator.validate(v) for v in value]
        return value, True

    def convert_to_proxy(self, value: str, **kwargs) -> UUID:
        raise NotImplementedError

    def convert_to_entity(self, value: list, **kwargs) -> str:
        if self.element_validator is not None:
            newval = [self.element_validator.convert_to_entity(v) for v in value]
            return newval

        raise NotImplementedError

    def repr_type(self):
        return self.ref_typename


class RefList(Reference):
    """A proxy property that manages a sub-schema (a sub-collection of other properties)"""

    VALIDATOR_CLASS = RefListField

    def __init__(self, *args, element_validator=UUIDField, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(element_validator, type):
            element_validator = element_validator()
        self.validator.element_validator = element_validator

    @classmethod
    def property_type_name(cls, proxy_type) -> str:
        ptnb = super().property_type_name(proxy_type)
        return f"List[{ptnb}]"

    def get(self, obj):
        """Low-level getter"""
        if obj.proxy_attr is None:
            obj.proxy_sync()
        return obj.proxy_attr[self.name]

    def __get__(self, obj, obj_type=None):
        if obj.proxy_attr is None:
            obj.proxy_sync()
        return ProxySublist(self, obj)

    def set(self, obj, value):
        raise NotImplementedError("Cannot set value of RefList field")

    def convert_entity_to_proxy(self, proxy, entity, **kwargs):
        entities = entity[self.entity_name]
        # entities is a list of entities, we need to fetch them from
        # our proxy's client.
        entity_id_name = self.proxy_type.proxy_schema.id.entity_name
        proxy_ids = [UUID(e.get(entity_id_name)) for e in entities]
        proxy.proxy_attr[self.name] = proxy_ids

    def convert_proxy_to_entity(self, proxy, entity, **kwargs):
        proxy_value = proxy.proxy_attr[self.name]
        if proxy_value is ...:
            return
        entity[self.entity_name] = self.validator.convert_to_entity(
            proxy_value, **kwargs
        )



import geojson

from .proxy.fieldvalidation import AnyField


class GeoJSON(AnyField):
    """
    A field for GeoJSON data.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(repr_type="GeoJSON", *args, **kwargs)
        self.add_check(self.to_geojson, 5)

    def to_geojson(self, value):
        try:
            value = geojson.GeoJSON.to_instance(value, strict=True)
        except ValueError:
            raise
        except Exception as e:
            raise ValueError("Invalid format for GeoJSON object") from e
        if not isinstance(value, geojson.GeoJSON):
            raise ValueError(f"Invalid object for GeoJSON: {value}")
        if not value.is_valid:
            raise ValueError(f"Invalid GeoJSON object: {value.errors()}")
        return value, False

    def convert_to_proxy(self, value, **kwargs):
        if value is None:
            return None
        return geojson.GeoJSON.to_instance(value)

    def convert_to_entity(self, value, **kwargs):
        return value

    def default_value(self):
        if hasattr(self, "default"):
            return self.default
        elif self.nullable:
            return None
        else:
            raise NotImplementedError()

    def repr_type(self):
        return self._repr_type



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

from .generic import GenericCursor, GenericProxy, api_call
from .proxy import (
    Id,
    NameId,
    ProxyList,
    Reference,
    RefList,
    Registry,
    RegistryCatalog,
    TagNameField,
    VocabNameField,
    derived_property,
)
from .utils import client_for, tag_split, validate_tagname

if TYPE_CHECKING:
    from .client import Client
    from .dataset import Dataset
    from .proxy.typing import TagSpecList


class Vocabulary(GenericProxy):
    """Vocabulary proxy provides manipulation of tag vocabularies."""

    id = Id()
    name = NameId(validator=VocabNameField)
    tags = RefList("Tag")

    @property
    def tagnames(self) -> list[str]:
        """The list of tag names."""
        return [t for t in self.tag_map]

    @property
    def tagspecs(self) -> TagSpecList:
        """The tagspecs of all tags in this vocabulary"""
        name = self.name
        return [f"{name}:{tagname}" for tagname in self.tagnames]

    @derived_property
    def tag_map(self, entity):
        """The tag_map is a dict mapping tag names to tag entities.

        This map is more conveniently accessed via the __getitem__ method
        on the vocabulary object.
        """
        return {t["name"]: t for t in entity["tags"]}

    def __getitem__(self, tagname: str) -> Tag:
        eid = UUID(self.tag_map[tagname]["id"])
        r = client_for(self).registry_for(Tag)
        return r.fetch_proxy(eid)

    def add_tags(self, taglist: list[str]):
        """Add tags to this vocabulary

        Args:
           taglist (list[str]): the list of tags to add.
        """
        r = client_for(self).registry_for(Tag)
        for tagname in taglist:
            Tag.new(r, name=tagname, vocabulary=self)

    # Overload to refresh cache
    def proxy_sync(self, entity=None):
        try:
            return super().proxy_sync(entity=entity)
        finally:
            self.proxy_registry.catalog.vocabulary_index.dirty = True

    @classmethod
    def new(
        cls,
        regspec: Registry | RegistryCatalog,
        *,
        name: str,
        tags: list[str] = [],
        autosync=True,
    ) -> Vocabulary:
        """Create a new vocabulary.

        Args:
            name (str): the name of the vocabulary.
            tags (list[str]): the list of tags to add to the vocabulary.
            autosync (bool): this is actually ignored.

        Returns:
            Vocabulary: the newly created vocabulary.
        """

        if not hasattr(cls, "proxy_schema"):
            raise TypeError(f"Class {cls.__name__} is not an entity class")

        tags = list(tags)
        if not all(validate_tagname(t) for t in tags):
            raise ValueError("Invalid tag name(s)", tags)

        newvoc = super().new(regspec, autosync=False, name=name, tags=[])

        ac = api_call(newvoc)
        ent = ac.vocabulary_create(name=name, tags=[{"name": t} for t in tags])

        newvoc.proxy_registry.register_proxy_for_entity(newvoc, ent)
        newvoc.proxy_from_entity(ent)
        newvoc.proxy_changed = None

        return newvoc


class Tag(GenericProxy):
    """A proxy for vocaublary tags.

    There is not much interesting functionality, since tags are not updatable.
    However, this proxy class allows for creation and deletion of vocabulary
    tags.

    Note that free tags (those without a vocabulary) in the Data Catalog are managed
    by the underlying CKAN implementation automatically (as they appear in tag fields
    for datasets, groups, organizations, etc.)
    """

    id = Id()
    name = NameId(validator=TagNameField)
    vocabulary = Reference(
        Vocabulary, nullable=False, entity_name="vocabulary_id", trigger_sync=True
    )

    @derived_property
    def tagspec(self, entity):
        tname = entity["name"]
        vid = entity["vocabulary_id"]
        if vid is not None:
            vname = self.proxy_registry.catalog.vocabulary_index.id_to_name[vid]
            return f"{vname}:{tname}"
        else:
            return tname

    def get_tagged_datasets(self) -> ProxyList[Dataset]:
        """Retrieve a number of datasets tagged with this tag.

        Note that there is an upper limit to the number of datasets
        (currently, 1000). For a more flexible access, the dataset search
        facility (Client.datasets.with_tag()) can be used.

        However, this call is convenient for 'rare' tags.
        """
        # Since tags are immutable, ignore own state
        c = client_for(self)
        return c.datasets.with_tag(self.tagspec)


class VocabularyCursor(GenericCursor[Vocabulary]):
    """Implement CKAN cursor functionalities for Vocabulary.

    N.B. Currently, this class is here as a placeholder and may eventually be
    removed.
    """

    def __init__(self, client):
        super().__init__(client, Vocabulary)


class TagCursor(GenericCursor[Tag]):
    """Tag cursors are a bit different, since they need to cater to
    free tags as well as vocabulary tags, and be fast about searching
    tags.

    The cursor fetches only retrieve free tags. However, there are other
    facilities that allow for the retrieval of all types of tags, as well
    as search operations using them.
    """

    def __init__(self, client: Client):
        super().__init__(client, Tag)

    def __getitem__(self, tagspec):
        if isinstance(tagspec, slice):
            return super().__getitem__(tagspec)

        ac = api_call(self)
        match tag_split(tagspec):
            case (None, tagname):
                entity = ac.tag_show(id=tagname)
            case (vocname, tagname):
                # Lookup the tag in the vocabulary index
                try:
                    self.client.vocabulary_index.dirty = True
                    entity = self.client.vocabulary_index.name_to_tags[vocname][tagname]
                except KeyError:
                    raise ValueError(f"Tag {tagspec} not found")
                # entity = ac.tag_show(id=tagname, vocabulary_id=vocname)
            case _:
                raise RuntimeError("Tag splitting, this is a bug")
        return self.client.registry_for(Tag).fetch_proxy_for_entity(entity)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any
from uuid import UUID

from .generic import GenericCursor, GenericProxy, generic_proxy_sync
from .mutils import is_s3url, s3spec_to_url
from .pdutils import read_dataframe
from .proxy import (
    DateField,
    ExtrasProperty,
    Id,
    IntField,
    Property,
    Proxy,
    ProxyVec,
    Reference,
    StrField,
)
from .proxy.refs import RefField
from .reprstyle import resource_to_html
from .utils import client_for

if TYPE_CHECKING:
    from .mutils import S3ObjSpec


class ExtrasResourceProperty(ExtrasProperty):
    """The handling of extras is different for Resource entities."""

    # The format is that extra attributes appear normally in
    # resource entities; there is no extras attribute.

    def convert_entity_to_proxy(self, proxy: Proxy, entity: Any, **kwargs):
        # to recognize extras fields, use the schema
        all_fields = proxy.proxy_schema.all_entity_fields
        proxy_extras = {p: v for p, v in entity.items() if p not in all_fields}
        proxy.proxy_attr[self.name] = proxy_extras

    def convert_proxy_to_entity(self, proxy: Proxy, entity: dict, **kwargs):
        proxy_extras = proxy.proxy_attr[self.name]
        entity |= proxy_extras

    def convert_to_create(
        self, proxy_type, create_props: dict, entity_props: dict, **kwargs
    ):
        all_fields = proxy_type.proxy_schema.all_fields

        # Collect all entries that do not appear in the schema
        entity_extras = {p: v for p, v in create_props.items() if p not in all_fields}
        entity_props |= entity_extras


class PackageField(RefField):
    """A field validator for package references."""

    @property
    def proxy_type(self):
        """Return the proxy type for the package reference."""
        return self.ref_property.proxy_types


class PackageRef(Reference):
    """A polymprphic reference to a package entity, used in Resource."""

    VALIDATOR_CLASS = PackageField

    # All entities with resources
    ENTITIES = ("Dataset", "Process", "Tool", "Workflow")

    def __init__(self):
        super().__init__(
            proxy_type="Dataset",
            nullable=False,
            trigger_sync=True,
            entity_name="package_id",
            updatable=False,
        )

    @cached_property
    def proxy_types(self):
        """Return the proxy types that this reference can point to."""
        from .proxy.schema import Schema

        return tuple(Schema.for_entity(t).cls for t in self.ENTITIES)

    def registry_for(self, obj) -> Any:
        """Return the registry for the package reference, given owner."""
        # Use the package_type attribute to determine the type of package.
        if (
            obj.proxy_attr
            and "_extras" in obj.proxy_attr
            and "package_type" in obj.proxy_attr["_extras"]
        ):
            package_type = obj.proxy_attr["_extras"].get("package_type", "dataset")
        else:
            package_type = "dataset"

        return obj.proxy_registry.catalog.registry_for(package_type.capitalize())


class Resource(GenericProxy):
    """
    A proxy for a STELAR resource with metadata and additional details.
    """

    id = Id()

    # dataset = Reference("Dataset", entity_name="package_id", trigger_sync=True)
    dataset = PackageRef()

    position = Property(validator=IntField)
    state = Property(validator=StrField)
    metadata_modified = Property(validator=DateField)

    url = Property(validator=StrField, updatable=True)
    format = Property(validator=StrField, updatable=True)
    description = Property(validator=StrField, updatable=True)
    hash = Property(validator=StrField, updatable=True)

    name = Property(validator=StrField, updatable=True)
    resource_type = Property(validator=StrField, updatable=True)
    mimetype = Property(validator=StrField, updatable=True)
    mimetype_inner = Property(validator=StrField, updatable=True)
    cache_url = Property(validator=StrField, updatable=True)
    size = Property(validator=IntField, updatable=True)
    created = Property(validator=DateField, updatable=True)
    last_modified = Property(validator=DateField, updatable=True)
    cache_last_updated = Property(validator=DateField, updatable=True)
    url_type = Property(validator=StrField, updatable=True)

    _extras = ExtrasResourceProperty()

    def __getattr__(self, attr):
        try:
            return self.proxy_schema.extras.get(self)[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __setattr__(self, attr, value):
        if attr.startswith("proxy_") or attr in self.proxy_schema.all_fields:
            return object.__setattr__(self, attr, value)

        prop = self.proxy_schema.extras

        # TODO: value validation: It is not clear what to do, presumably the correct
        # value would be transformable to json

        prop.touch(self)
        self.proxy_schema.extras.get(self)[attr] = value
        self.proxy_autocommit()

    def __delattr__(self, attr):
        if attr.startswith("proxy_") or attr in self.proxy_schema.all_fields:
            return object.__delattr__(self, attr)

        prop = self.proxy_schema.extras
        extras = prop.get(self)
        if attr in extras:
            prop.touch(self)
            del extras[attr]
            self.proxy_autocommit()
        else:
            raise AttributeError(attr)

    def proxy_sync(self, entity=None):
        return generic_proxy_sync(self, entity, update_method="update")

    def __str__(self):
        """
        Provide a human-readable string representation of the Resource instance.

        Returns:
            str: A string describing the resource's key attributes.
        """
        return f"Resource ID: {self.id} | Name: {self.name} | URL: {self.url} | Format : {self.format}"

    def _repr_html_(self):
        """
        Provide an HTML representation of the Resource instance for Jupyter display,
        with enhanced styles, watermark, and consistent formatting.
        """
        return resource_to_html(self)

    def open(self, mode="rb", **kwargs):
        """
        Open the resource for reading or writing.

        Args:
            mode (str): The mode in which to open the resource. This can be one of the
                following: 'r', 'rb', 'w', 'wb', 'a', 'ab'.

        Returns:
            file-like: A file-like object that can be used to read or write data.
        """
        client = client_for(self)
        url = self.url
        if not is_s3url(url):
            raise ValueError("Only s3 URLs are supported for path")
        return client.s3fs_open(url, mode=mode, **kwargs)

    def read_dataframe(self, format=None, **kwargs):
        """
        Read a DataFrame from the resource.

        Args:
            format (str): The format of the file to read. If not specified, the format will be
                inferred from the file extension.
            kwargs (dict): Additional keyword arguments to pass to the read.

        Returns:
            pd.DataFrame: The DataFrame read from the resource.
        """
        client = client_for(self)
        format = format or self.format or None
        return read_dataframe(client, self.url, format=format, **kwargs)


class ResourceCursor(GenericCursor):
    """
    A cursor for a collection of STELAR resources.
    """

    def __init__(self, api):
        super().__init__(api, Resource)

    def fetch(self, **kwargs):
        raise NotImplementedError("ResourceCursor does not support fetch operations.")

    def fetch_list(self, **kwargs):
        raise NotImplementedError("ResourceCursor does not support fetch operations.")

    def search_url(self, path: str) -> ProxyVec:
        """Return resources whose url matches the given string.

        Args:
            path (str): The partial URL to search for.
        Returns:
            ProxyVec: A vector of resources with matching URLs.
        """
        res = self.search(query=[f"url:{path}"])
        rl = [UUID(r["id"]) for r in res["results"]]
        return ProxyVec(self.client, Resource, rl)

    def for_object(self, s3obj_spec: S3ObjSpec) -> ProxyVec:
        """Return resources for a given S3 object specification.

        Args:
            s3obj_spec (S3ObjSpec): The S3 object specification.
        Returns:
            ProxyVec: A vector of resources for the given S3 object
        """
        s3url = s3spec_to_url(s3obj_spec)
        return self.search_path(s3url)

    def search(
        self,
        *,
        query: list[str] | None = [],
        order_by: str = None,
        limit: int | None = None,
        offset: int | None = None,
    ):
        """
        Search for resources.

        This is the main function for searching the data catalog for resources.
        Other functions are implemented on top of this one.

        A resource query is a conunction of query terms of the form <field>:<match>.
        The query terms are combined with AND. The match is a string with optional wildcards.


        Arguments:
            query: A list of query strings. Each query string is of the form "<field>:<text>".
            order_by: A field to order by. Only a single field is allowed and only ascending
                     order is supported.
            limit: The maximum number of results to return.
            offset: The offset to start from.

        Returns:
            An answer dict which contains the following fields:
            - count: The number of results found (not the number of results returned).
            - results: A list of dicts, each corresponding to a resource entity.

        """
        search = self.client.api.get_call(Resource, "search")
        query = dict(
            query=query,
            order_by=order_by,
            limit=limit,
            offset=offset,
        )
        return search({k: v for k, v in query.items() if v is not None})
